"""HTTP Transport  for asyncio."""
import asyncio
import logging
from binascii import unhexlify
from contextlib import suppress

import aiohttp

from .http_reader_writer import HttpReaderWriter
from .msg_to_url import convert_to_url
from .hub_connection_exception import HubConnectionException

_LOGGER = logging.getLogger(__name__)
READ_WAIT = 0.5
SESSION_RETRIES = 30
RECONNECT_WAIT = 7


async def async_connect_http(host, username, password, protocol, port=None):
    """Connect to the Hub Version 2 via HTTP."""
    port = 25105 if not port else port
    transport = HttpTransport(
        protocol=protocol, host=host, port=port, username=username, password=password
    )
    if await transport.async_test_connection():
        transport.start_reader()
        protocol.connection_made(transport)
        return transport
    return None


# This transport is designed for the Hub version 2.
# Hub version 1 should use socket interface on port 9761.
# pylint: disable=too-many-instance-attributes
class HttpTransport(asyncio.Transport):
    """An asyncio transport model of an HTTP communication channel.

    A transport class is an abstraction of a communication channel.
    This allows protocol implementations to be developed against the
    transport abstraction without needing to know the details of the
    underlying channel, such as whether it is a pipe, a socket, or
    indeed an HTTP connection.


    You generally won’t instantiate a transport yourself; instead, you
    will call `create_http_connection` which will create the
    transport and try to initiate the underlying communication channel,
    calling you back when it succeeds.
    """

    def __init__(self, protocol, host, username, password, port=25105):
        """Init the HttpTransport class."""
        super().__init__()
        self._protocol = protocol
        self._host = host
        self._port = port
        self._auth = aiohttp.BasicAuth(username, password)
        self._reader_writer = HttpReaderWriter(self._auth)

        self._closing = False
        self._read_write_lock = asyncio.Lock()
        self._last_read = asyncio.Queue()
        self._last_msg = None
        self._reader_task = None

    def abort(self):
        """Alternative to closing the transport."""
        self.close()

    def can_write_eof(self):
        """Return False always."""
        return False

    def is_closing(self):
        """Return the state of the transport.

        True if the transport is closed or in the process of closing.

        """
        return self._closing

    def close(self):
        """Close the transport."""
        _LOGGER.debug("Closing Hub session")
        self._closing = True

    def get_write_buffer_size(self):
        """Rreturn 0 (i.e. none) always."""
        return 0

    def pause_reading(self):
        """Pause the read."""
        asyncio.ensure_future(self._stop_reader(False))

    def resume_reading(self):
        """Resume the reader."""
        self._start_reader()

    def set_write_buffer_limits(self, high=None, low=None):
        """Not implemented."""
        raise NotImplementedError("HTTP connections do not support write buffer limits")

    def write(self, data):
        """Write data to the transport."""
        asyncio.ensure_future(self.async_write(data))

    async def async_write(self, data):
        """Async write to the transport."""
        url = convert_to_url(self._host, self._port, data)
        await self._async_write_url(url=url, msg=bytes(data))

    def start_reader(self):
        """Start the reader."""
        if self._reader_task is None:
            self._start_reader()

    async def _async_write_url(self, url, msg=None):
        """Write the message to the Hub."""
        await self._read_write_lock
        response_status = 0
        retry = 0
        while response_status != 200 and retry < 5:
            if self.is_closing():
                if self._read_write_lock.locked():
                    self._read_write_lock.release()
                return
            response_status = await self._reader_writer.async_write(url)
            if response_status != 200:
                _LOGGER.debug("Hub write request failed for url %s", url)
                retry += 1
                await asyncio.sleep(READ_WAIT)
            else:
                self._last_msg = msg
        if self._read_write_lock.locked():
            self._read_write_lock.release()

    async def async_test_connection(self):
        """Test the connection to the hub."""
        url = "http://{:s}:{:d}/buffstatus.xml".format(self._host, self._port)
        response = await self._reader_writer.async_test_connection(url)
        if not response:
            self.close()
        return response

    def write_eof(self):
        """Not implemented."""
        raise NotImplementedError("HTTP connections do not support end-of-file")

    def writelines(self, list_of_data):
        """Not implemented."""
        raise NotImplementedError("HTTP connections do not support writelines")

    async def _clear_buffer(self):
        _LOGGER.debug("..................Clearing the buffer..............")
        url = "http://{:s}:{:d}/1?XB=M=1".format(self._host, self._port)
        await self._async_write_url(url)

    # pylint: disable=broad-except
    async def _ensure_reader(self):
        _LOGGER.info("Insteon Hub reader started")
        await self._clear_buffer()
        await self._reader_writer.reset_reader()
        url = "http://{:s}:{:d}/buffstatus.xml".format(self._host, self._port)
        retry = 0
        while not self._closing:
            buffer = None
            await self._read_write_lock
            try:
                buffer = await self._reader_writer.async_read(url)
            except HubConnectionException:
                retry += 1
                _LOGGER.debug("Connection retry: %d", retry)
                await asyncio.sleep(RECONNECT_WAIT)
                if retry >= SESSION_RETRIES:
                    _LOGGER.error(
                        "Closing connection Hub after %d retries", SESSION_RETRIES
                    )
                    self.close()

            except (asyncio.CancelledError, GeneratorExit) as ex:
                _LOGGER.debug("Stop connection to Hub: %s", ex)
                self.close()

            else:
                retry = 0
                if buffer:
                    _LOGGER.debug("New buffer: %s", buffer)
                    buffer = self._check_strong_nak(buffer)
                    bin_buffer = unhexlify(buffer)
                    self._protocol.data_received(bin_buffer)
            if self._read_write_lock.locked():
                self._read_write_lock.release()
            if not self.is_closing:
                await asyncio.sleep(READ_WAIT)
        _LOGGER.info("Insteon Hub reader stopped")

    def _check_strong_nak(self, buffer):
        """Check if a NAK message is received with multiple NAKs.

        There appears to be a bug in the Hub that produces a series
        of NAK codes rather than returning the original message
        with a single NAK.
        """
        naks = buffer.count("15")
        if (naks >= 9 or naks == len(buffer) / 2) and self._last_msg is not None:
            return self._last_msg.hex() + "15"
        return buffer

    # pylint: disable=unused-argument
    def _start_reader(self, future=None):
        _LOGGER.debug("Starting the buffer reader")
        if not self._closing:
            self._reader_task = asyncio.ensure_future(self._ensure_reader())
            self._reader_task.add_done_callback(self._protocol.connection_lost)

    async def _stop_reader(self, reconnect=False):
        _LOGGER.debug("Stopping the reader and reconnect is %s", reconnect)
        self.close()
        if self._reader_task:
            if not reconnect:
                self._reader_task.remove_done_callback(self._protocol.connection_lost)
            self._reader_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._reader_task
                await asyncio.sleep(0)



"""All-Link Record Flags."""

from binascii import hexlify

from ...utils import bit_is_set, set_bit


def create(auto_link: bool, monitor_mode: bool, auto_led: bool, disable_deadman: bool):
    """Create an IM Configuration Flag entity."""
    flags = IMConfigurationFlags(0x00)
    flags.is_auto_link = auto_link
    flags.is_monitor_mode = monitor_mode
    flags.is_auto_led = auto_led
    flags.is_disable_deadman = disable_deadman
    return flags


def create_template(
    auto_link: bool = None,
    monitor_mode: bool = None,
    auto_led: bool = None,
    disable_deadman: bool = None,
):
    """Create an IM Configuration Flag entity."""
    flags = IMConfigurationFlags(0x00)
    flags.is_auto_link = auto_link
    flags.is_monitor_mode = monitor_mode
    flags.is_auto_led = auto_led
    flags.is_disable_deadman = disable_deadman
    return flags


def _normalize(data):
    if isinstance(data, IMConfigurationFlags):
        return bytes(data)
    return data


class IMConfigurationFlags:
    """IM Configuration Flags."""

    def __init__(self, data: int):
        """Init the IMConfigurationFlags class."""
        data = _normalize(data)
        if isinstance(data, bytes):
            data = int.from_bytes(data, byteorder="big")
        self._auto_link = bit_is_set(data, 7)
        self._monitor_mode = bit_is_set(data, 6)
        self._auto_led = bit_is_set(data, 5)
        self._disable_deadman = bit_is_set(data, 4)

    def __bytes__(self):
        """Return the byte representation of the flags."""
        flags = set_bit(0x00, 7, self._auto_link)
        flags = set_bit(flags, 6, self._monitor_mode)
        flags = set_bit(flags, 5, self._auto_led)
        flags = set_bit(flags, 4, self._disable_deadman)
        return bytes([flags])

    def __repr__(self):
        """Return the hex representation of the flags."""
        val = {
            "auto link": 1 if self._auto_link else 0,
            "monitor mode": 1 if self._monitor_mode else 0,
            "auto led": 1 if self._auto_led else 0,
            "disable deadman": 1 if self._disable_deadman else 0,
        }
        return str(val)

    def __str__(self):
        """Return the hex representation of the flags."""
        return hexlify(bytes(self)).decode()

    def __eq__(self, other):
        """Check if this is equal to other."""
        if not isinstance(other, IMConfigurationFlags):
            return False
        return (
            self.is_auto_link == other.is_auto_link
            and self.is_monitor_mode == other.is_monitor_mode
            and self.is_auto_led == other.is_auto_led
            and self.is_disable_deadman == other.is_disable_deadman
        )

    def __hash__(self):
        """Represent the IMConfigurationFlags class as a hash."""
        return hash(bytes(self))

    @property
    def is_auto_link(self):
        """Return if record is in use."""
        return self._auto_link

    @is_auto_link.setter
    def is_auto_link(self, val: bool):
        """Set the auto link value."""
        if val is None:
            self._auto_link = None
        else:
            self._auto_link = bool(val)

    @property
    def is_monitor_mode(self):
        """Return if the record is a responder or controller."""
        return self._monitor_mode

    @is_monitor_mode.setter
    def is_monitor_mode(self, val: bool):
        """Set the monitoring mode value."""
        if val is None:
            self._monitor_mode = None
        else:
            self._monitor_mode = bool(val)

    @property
    def is_auto_led(self):
        """Return if the record is the high water mark."""
        return self._auto_led

    @is_auto_led.setter
    def is_auto_led(self, val: bool):
        """Set the auto LED value."""
        if val is None:
            self._auto_led = None
        else:
            self._auto_led = bool(val)

    @property
    def is_disable_deadman(self):
        """Return if disable deadman is set."""
        return self._disable_deadman

    @is_disable_deadman.setter
    def is_disable_deadman(self, val: bool):
        """Set the disable deadman value."""
        if val is None:
            self._disable_deadman = None
        else:
            self._disable_deadman = bool(val)



"""Base device object."""
from inspect import getfullargspec
import logging
from abc import ABC
import asyncio
from datetime import datetime

from ..address import Address
from ..aldb import ALDB
from ..constants import EngineVersion
from ..default_link import DefaultLink
from ..handlers.to_device.product_data_request import ProductDataRequestCommand
from ..handlers.to_device.engine_version_request import EngineVersionRequest
from ..managers.get_set_ext_property_manager import GetSetExtendedPropertyManager
from ..managers.get_set_op_flag_manager import GetSetOperatingFlagsManager
from ..managers.link_manager.default_links import async_add_default_links
from ..operating_flag import OperatingFlag
from ..utils import multiple_status

_LOGGER = logging.getLogger(__name__)


class Device(ABC):
    """INSTEON Device Class."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the Device class."""
        self._address = Address(address)
        self._cat = cat
        self._subcat = subcat
        if self._subcat is None:
            self._subcat = 0x00
        self._firmware = firmware if firmware is not None else 0x00
        self._description = description
        self._model = model
        self._product_id = None
        self._is_battery = False
        self._engine_version = EngineVersion.UNKNOWN

        self._last_communication_received = datetime(1, 1, 1, 1, 1, 1)
        self._product_data_in_aldb = False
        self._groups = {}
        self._handlers = {}
        self._managers = {}
        self._events = {}

        self._aldb = ALDB(self._address)
        self._default_links = []
        self._operating_flags = {}
        self._properties = {}
        self._op_flags_manager = GetSetOperatingFlagsManager(
            self._address, self._operating_flags
        )
        self._ext_property_manager = GetSetExtendedPropertyManager(self._address)

        self._register_operating_flags()
        self._register_groups()
        self._register_events()
        self._register_handlers_and_managers()
        self._subscribe_to_handelers_and_managers()
        self._register_default_links()

    @property
    def address(self):
        """Return the INSTEON device address."""
        return self._address

    @property
    def cat(self):
        """Return the INSTEON device category."""
        return self._cat

    @property
    def subcat(self):
        """Return the INSTEON device subcategory."""
        return self._subcat

    @property
    def firmware(self):
        """Return the INSTEON product key."""
        return self._firmware

    @property
    def description(self):
        """Return the INSTEON device description."""
        return self._description

    @property
    def model(self):
        """Return the INSTEON device model number."""
        return self._model

    @property
    def product_id(self):
        """Return the INSTEON product ID."""
        return self._product_id

    @property
    def id(self):
        """Return the ID of the device."""
        return self._address.id

    @property
    def groups(self):
        """Return the device groups / buttons."""
        return self._groups

    @property
    def events(self):
        """Return the device events."""
        return self._events

    @property
    def prod_data_in_aldb(self):
        """Return if the PLM use the ALDB data to setup the device.

        True if Product data (cat, subcat) is stored in the PLM ALDB.
        False if product data must be acquired via a Device ID message or from
        a Product Data Request command.

        The method of linking determines if product data in the ALDB,
        therefore False is the default. The common reason to store product data
        in the ALDB is for one way devices or battery opperated devices where
        the ability to send a command request is limited.
        """
        return self._product_data_in_aldb

    @property
    def aldb(self):
        """Return the device All-Link Database."""
        return self._aldb

    @property
    def operating_flags(self):
        """Return the Operating Flags."""
        return self._operating_flags

    @property
    def properties(self):
        """Return the Extended Properties."""
        return self._properties

    @property
    def default_links(self):
        """Return the list of default links."""
        return self._default_links

    @property
    def is_battery(self):
        """Return True if the device is battery operated."""
        return self._is_battery

    @property
    def engine_version(self):
        """Return device engine version."""
        return self._engine_version

    @engine_version.setter
    def engine_version(self, value: EngineVersion):
        """Set the device engine version."""
        try:
            version = EngineVersion(value)
        except ValueError:
            version = EngineVersion.UNKNOWN
        if version in [EngineVersion.I2, EngineVersion.I2CS]:
            self._op_flags_manager.extended_write = True
        self._engine_version = version

    def status(self, group=None):
        """Get the status of the device."""
        args = getfullargspec(self.async_status)
        if "group" in args[0]:
            asyncio.ensure_future(self.async_status(group=group))
        else:
            asyncio.ensure_future(self.async_status())

    async def async_status(self, group=None):
        """Get the status of the device."""

    async def async_read_config(self):
        """Get all configuration settings.

        This includes:
        - Operating flags
        - Extended properties
        - All-Link Database records.
        """
        result_op_flags = await self.async_read_op_flags()
        result_ext_prop = await self.async_read_ext_properties()
        result_aldb = await self._aldb.async_load()
        return multiple_status(result_ext_prop, result_op_flags), result_aldb

    async def async_read_op_flags(self, group=None):
        """Read the device operating flags."""
        return await self._op_flags_manager.async_read(group=group)

    async def async_write_op_flags(self):
        """Write the operating flags to the device."""
        return await self._op_flags_manager.async_write()

    async def async_read_ext_properties(self, group=None):
        """Get the device extended properties."""
        return await self._ext_property_manager.async_read(group=group)

    async def async_write_ext_properties(self):
        """Write the extended properties."""
        return await self._ext_property_manager.async_write()

    async def async_read_product_id(self):
        """Get the product ID."""
        return await self._handlers["product_data_cmd"].async_send()

    async def async_add_default_links(self):
        """Add the default links betweent he modem and the device."""
        return await async_add_default_links(self)

    async def async_get_engine_version(self):
        """Read the device engine version."""
        return await self._handlers["engine_version_cmd"].async_send()

    def _register_groups(self):
        """Add the groups to the device."""

    def _register_handlers_and_managers(self):
        """Add all handlers to the device and register listeners."""
        self._handlers["product_data_cmd"] = ProductDataRequestCommand(self._address)
        self._handlers["engine_version_cmd"] = EngineVersionRequest(self._address)

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe groups and events to handlers and managers."""
        self._handlers["product_data_cmd"] = ProductDataRequestCommand(self._address)
        self._handlers["engine_version_cmd"].subscribe(self._engine_version_received)

    def _register_default_links(self):
        """Add default links for linking the device to the modem."""
        link_0 = DefaultLink(
            is_controller=False,
            group=0,
            dev_data1=0,
            dev_data2=0,
            dev_data3=0,
            modem_data1=self.cat,
            modem_data2=self.subcat,
            modem_data3=self.firmware,
        )
        link_0 = self._default_links.append(link_0)

    def _register_events(self):
        """Add events that are triggered when events are fired from the device."""

    def _register_operating_flags(self):
        """Add operating flags to the device."""

    def _add_operating_flag(
        self, name, group, bit, set_cmd, unset_cmd, is_reversed=False
    ):
        flag_type = bool if bit is not None else int
        self._operating_flags[name] = OperatingFlag(
            self._address, name, flag_type, is_reversed
        )
        self._op_flags_manager.subscribe(name, group, bit, set_cmd, unset_cmd)

    def _add_property(
        self, name, data_field, set_cmd, group=1, bit=None, is_reversed=False
    ):
        self._properties[name] = self._ext_property_manager.create(
            name, group, data_field, bit, set_cmd, is_reversed
        )

    def _remove_operating_flag(self, name, group=None):
        self._op_flags_manager.unsubscribe(name)
        self._operating_flags.pop(name)

    def _handle_product_data(self, product_id, cat, subcat):
        """Receive and set the product data information."""
        self._product_id = product_id

    def _engine_version_received(self, engine_version):
        """Receive engine version response."""
        self.engine_version = engine_version

    def _product_data_received(self, product_id, cat, subcat):
        """Receive product data response."""
        self._product_id = product_id
        self._cat = cat
        self._subcat = subcat



"""All-Link database for an Insteon Modem."""
import logging
from typing import Callable

from . import ALDBBase
from .. import pub
from ..address import Address
from ..constants import ALDBStatus, ALDBVersion, ManageAllLinkRecordAction
from ..handlers import ResponseStatus
from ..handlers.manage_all_link_record import ManageAllLinkRecordCommand
from ..managers.im_read_manager import ImReadManager
from ..topics import ALL_LINK_RECORD_RESPONSE
from .aldb_record import ALDBRecord

_LOGGER = logging.getLogger(__name__)
MAX_RETRIES = 3


class ModemALDB(ALDBBase):
    """All-Link database for modems.

    Subscribed topics:
    modem.aldb.loaded: Triggered when the ALDB load command completes.

    Messages sent:
    modem.aldb.load: Triggers the loading of the ALDB.
    """

    def __init__(self, address, version=ALDBVersion.V2, mem_addr=0x1FFF):
        """Init the ModemALDB."""

        super().__init__(address, version, mem_addr)

        # If we are not the first modem, don't subscribe to
        mgr = pub.getDefaultTopicMgr()
        topic = mgr.getTopic(ALL_LINK_RECORD_RESPONSE, okIfNone=True)
        if not topic:
            self._read_manager = ImReadManager(self)
        else:
            self._read_manager = None

        self._write_cmd = ManageAllLinkRecordCommand()

    def __setitem__(self, mem_addr, record):
        """Add or Update a device in the ALDB."""
        if not isinstance(record, ALDBRecord):
            raise ValueError

        self._records[mem_addr] = record

    # pylint: disable=arguments-differ
    async def async_load(self, callback: Callable = None):
        """Load the All-Link Database."""
        _LOGGER.debug("Loading the modem ALDB")
        self._records = {}
        if self._read_manager is not None:
            await self._read_manager.async_load()
        self._status = ALDBStatus.LOADED
        if callback:
            callback()
        return self._status

    def add(
        self,
        group: int,
        target: Address,
        controller: bool,
        data1: int = 0,
        data2: int = 0,
        data3: int = 0,
    ):
        """Add a record to the All-Link database."""
        record = ALDBRecord(
            memory=0,
            in_use=True,
            controller=controller,
            high_water_mark=False,
            bit5=True,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
        )
        self._dirty_records.append(record)

    def remove(self, controller, group, target):
        """Remove a record to the All-Link database."""
        record = ALDBRecord(
            memory=0,
            in_use=False,
            controller=controller,
            high_water_mark=False,
            bit5=True,
            group=group,
            target=target,
            data1=0,
            data2=0,
            data3=0,
        )
        self._dirty_records.append(record)

    async def async_write(self):
        """Write modified records to the device.

        Returns a tuple of (completed, failed) record counts.
        """
        completed = []
        failed = []
        self._id_recs_to_restore()
        while self._dirty_records:
            rec = self._dirty_records.pop()
            if not rec.is_in_use:
                action = ManageAllLinkRecordAction.DELETE_FIRST
            elif rec.is_controller:
                action = ManageAllLinkRecordAction.MOD_FIRST_CTRL_OR_ADD
            else:
                action = ManageAllLinkRecordAction.MOD_FIRST_RESP_OR_ADD
            retries = 0
            response = ResponseStatus.UNSENT
            while response != ResponseStatus.SUCCESS and retries < MAX_RETRIES:
                response = await self._write_cmd.async_send(
                    action=action,
                    controller=rec.is_controller,
                    group=rec.group,
                    target=rec.target,
                    data1=rec.data1,
                    data2=rec.data2,
                    data3=rec.data3,
                    in_use=rec.is_in_use,
                    high_water_mark=rec.is_high_water_mark,
                    bit5=rec.is_bit5_set,
                    bit4=rec.is_bit4_set,
                )
                retries += 1
            if response == ResponseStatus.SUCCESS:
                completed.append(rec)
            else:
                failed.append(rec)

        return len(completed), len(failed)

    def _id_recs_to_restore(self):
        """Find matching records to delete and restore.

        When deleting a record it is required to delete all that match the target and group
        then restore all the ones that do not match the controller. (yes it is stupid)
        """
        deleted_recs = []
        restore_recs = []
        multi_delete = []
        for rec in self._dirty_records:
            if not rec.is_in_use:
                deleted_recs.append(rec)

        for del_rec in deleted_recs:
            found_restore = False
            for mem_addr in self._records:
                rec = self._records[mem_addr]
                if rec.target == del_rec.target and rec.group == del_rec.group:
                    new_rec = self._set_not_in_use(rec)
                    multi_delete.append(new_rec)
                    if (
                        not rec.is_controller == del_rec.is_controller
                        and not found_restore
                    ):
                        # We only need one controller or responder for a target and group
                        restore_recs.append(rec)
        self._dirty_records.extend(multi_delete)
        self._dirty_records.extend(restore_recs)

    @classmethod
    def _set_not_in_use(cls, rec):
        new_rec = ALDBRecord(
            memory=rec.mem_addr,
            controller=rec.is_controller,
            group=rec.group,
            target=rec.target,
            data1=rec.data1,
            data2=rec.data2,
            data3=rec.data3,
            in_use=False,
            high_water_mark=False,
            bit4=True,
            bit5=False,
        )
        return new_rec



"""Manage outbound ON All-Link Cleanup ACK response to a device."""

from ...topics import ON
from .all_link_cleanup_ack_command import AllLinkCleanupAckCommandHandlerBase


class OnAllLinkCleanupAckCommand(AllLinkCleanupAckCommandHandlerBase):
    """Manage outbound ON All-Link Cleanup ACK command to a device."""

    def __init__(self, address, group):
        """Init the OnLevelCommand class."""
        super().__init__(topic=ON, address=address, group=group)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the ON command async."""
        return await super().async_send(on_level=self._group)



"""Get Device Info command handler."""
from ...address import Address
from ...topics import ENTER_LINKING_MODE
from ..from_device.assign_to_all_link_group import AssignToAllLinkGroupCommand
from .direct_command import DirectCommandHandlerBase


class EnterLinkingModeCommand(DirectCommandHandlerBase):
    """Place a device in linking mode command handler."""

    def __init__(self, address: Address):
        """Init the IdRequest class."""
        super().__init__(topic=ENTER_LINKING_MODE, address=address)
        self._link_complete_handler = AssignToAllLinkGroupCommand(address)

    # pylint: disable=arguments-differ
    async def async_send(self, group: int = 0):
        """Send the device ID request asyncronously."""
        return await super().async_send(group=group)



"""Manage outbound OFF All-Link Broadcast command to a device."""

from ...topics import ON
from .all_link_broadcast_command import AllLinkBroadcastCommandHandlerBase


class OnLevelAllLinkBroadcastCommand(AllLinkBroadcastCommandHandlerBase):
    """Manage outbound OFF All-Link Broadcast command to a device."""

    def __init__(self, group):
        """Init the OnLevelCommand class."""
        super().__init__(topic=ON, group=group)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the ON command async."""
        return await super().async_send(group=0, on_level=0xFF)



"""Thermostat status manager."""
import asyncio
import logging
from collections import namedtuple

from ..address import Address
from ..constants import ResponseStatus
from ..handlers.from_device.thermostat_status_response import (
    ThermostatStatusResponseHandler,
)
from ..handlers.to_device.extended_get_2 import ExtendedGet2Command
from ..handlers.to_device.thermostat_get_set_point import ThermostatGetSetPointCommand
from ..handlers.from_device.thermostat_set_point_response import (
    ThermostatSetPointResponseHandler,
)
from ..utils import multiple_status

_LOGGER = logging.getLogger(__name__)
TIMEOUT = 5

PropertyInfo = namedtuple("PropertyInfo", "name group data_field bit set_cmd")


class GetThermostatStatus:
    """Thermostat status manager."""

    def __init__(self, address):
        """Init the GetThermostatStatus class."""
        self._address = Address(address)
        self._get_status_command = ExtendedGet2Command(self._address)
        self._get_set_point_command = ThermostatGetSetPointCommand(self._address)
        self._status_response = ThermostatStatusResponseHandler(self._address)
        self._set_point_response = ThermostatSetPointResponseHandler(self._address)
        self._response_status = asyncio.Queue()
        self._response_set_point = asyncio.Queue()
        self._status_response.subscribe(self._status_received)
        self._set_point_response.subscribe(self._set_point_received)

    async def async_status(self):
        """Read the device status."""
        status1 = await self._status()
        status2 = await self._set_point()
        return multiple_status(status1, status2)

    async def _status(self):
        """Send the status command and call set point command."""
        while not self._response_status.empty():
            self._response_status.get_nowait()
        retries = 3
        response_status = ResponseStatus.FAILURE
        while retries:
            await self._get_status_command.async_send()
            try:
                response_status = await asyncio.wait_for(
                    self._response_status.get(), TIMEOUT
                )
                return response_status
            except asyncio.TimeoutError:
                _LOGGER.error("Exception _status timed out, retries: %d", retries)
            retries -= 1
        return response_status

    async def _set_point(self):
        """Send the set point command."""
        while not self._response_set_point.empty():
            self._response_set_point.get_nowait()
        retries = 3
        response_set_point = ResponseStatus.FAILURE
        while retries:
            await self._get_set_point_command.async_send()
            try:
                return await asyncio.wait_for(self._response_set_point.get(), TIMEOUT)
            except asyncio.TimeoutError:
                _LOGGER.debug("Set point response timed out")
            retries -= 1
        return response_set_point

    def _status_received(
        self,
        day,
        hour,
        minute,
        second,
        system_mode,
        fan_mode,
        cool_set_point,
        humidity,
        temperature,
        cooling,
        heating,
        celsius,
        heat_set_point,
    ):
        """Notify the read process that the resonse was received."""
        self._response_status.put_nowait(ResponseStatus.SUCCESS)

    def _set_point_received(
        self,
        stage_1_on_minutes,
        humidity_high,
        humidity_low,
        firmwire,
        cool_set_point,
        heat_set_point,
        rf_offset,
    ):
        """Notify read process that the st point response was received."""
        self._response_set_point.put_nowait(ResponseStatus.SUCCESS)



"""Handle sending a read request for ALDB records."""
import logging

from .. import inbound_handler
from ...address import Address
from ...protocol.messages.user_data import UserData
from ...topics import EXTENDED_READ_WRITE_ALDB
from ...utils import build_topic
from ..inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


def _create_aldb_record_from_userdata(userdata: UserData):
    """Create ALDB Record from the userdata dictionary."""


class ReceiveALDBRecordHandler(InboundHandlerBase):
    """Receive an ALDB record direct inbound message."""

    def __init__(self, address: Address):
        """Init the ReceiveALDBRecordHandler class."""
        self._address = Address(address)
        super().__init__(topic=EXTENDED_READ_WRITE_ALDB, address=self._address)
        self._subscriber_topic = build_topic(
            prefix="handler.{}".format(self._address.id),
            topic="aldb_record_received",
            message_type="direct",
        )

    @inbound_handler
    def handle_response(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the inbound message."""
        _LOGGER.debug("ALDB Read direct message received")
        if user_data is not None and user_data.get("d2") == 0x01:
            memhi = user_data.get("d3")
            memlo = user_data.get("d4")
            memory = memhi << 8 | memlo
            control_flags = user_data.get("d6")
            if isinstance(control_flags, bytes):
                control_flags = int.from_bytes(control_flags, "big")
            in_use = bool(control_flags & 1 << 7)
            controller = bool(control_flags & 1 << 6)
            bit5 = bool(control_flags & 1 << 5)
            bit4 = bool(control_flags & 1 << 4)
            high_water_mark = not bool(control_flags & 1 << 1)
            group = user_data.get("d7")
            addrhi = user_data.get("d8")
            addrmed = user_data.get("d9")
            addrlo = user_data.get("d10")
            target = Address(bytearray([addrhi, addrmed, addrlo]))
            data1 = user_data.get("d11")
            data2 = user_data.get("d12")
            data3 = user_data.get("d13")
            self._call_subscribers(
                memory=memory,
                controller=controller,
                group=group,
                target=target,
                data1=data1,
                data2=data2,
                data3=data3,
                in_use=in_use,
                high_water_mark=high_water_mark,
                bit5=bit5,
                bit4=bit4,
            )



"""Constants used to ensure topic consistantancy."""

STANDARD_RECEIVED = "standard_received"
EXTENDED_RECEIVED = "extended_received"
X10_RECEIVED = "x10_received"
ALL_LINKING_COMPLETED = "all_linking_completed"
BUTTON_EVENT_REPORT = "button_event_report"
USER_RESET_DETECTED = "user_reset_detected"
ALL_LINK_CLEANUP_FAILURE_REPORT = "all_link_cleanup_failure_report"
ALL_LINK_RECORD_RESPONSE = "all_link_record_response"
ALL_LINK_CLEANUP_STATUS_REPORT = "all_link_cleanup_status_report"
GET_IM_INFO = "get_im_info"
SEND_ALL_LINK_COMMAND = "send_all_link_command"
SEND_STANDARD = "send_standard"
SEND_EXTENDED = "send_extended"
X10_SEND = "x10_send"
START_ALL_LINKING = "start_all_linking"
CANCEL_ALL_LINKING = "cancel_all_linking"
SET_HOST_DEV_CAT = "set_host_dev_cat"
RESET_IM = "reset_im"
SET_ACK_MESSAGE_BYTE = "set_ack_message_byte"
GET_FIRST_ALL_LINK_RECORD = "get_first_all_link_record"
GET_NEXT_ALL_LINK_RECORD = "get_next_all_link_record"
SET_IM_CONFIGURATION = "set_im_configuration"
GET_ALL_LINK_RECORD_FOR_SENDER = "get_all_link_record_for_sender"
LED_ON = "led_on"
LED_OFF = "led_off"
MANAGE_ALL_LINK_RECORD = "manage_all_link_record"
SET_NAK_MESSAGE_BYTE = "set_nak_message_byte"
SET_ACK_MESSAGE_TWO_BYTES = "set_ack_message_two_bytes"
RF_SLEEP = "rf_sleep"
GET_IM_CONFIGURATION = "get_im_configuration"
# Command Topics
ASSIGN_TO_ALL_LINK_GROUP = "assign_to_all_link_group"
DELETE_FROM_ALL_LINK_GROUP = "delete_from_all_link_group"
PRODUCT_DATA_REQUEST = "product_data_request"
FX_USERNAME = "fx_username"
DEVICE_TEXT_STRING_REQUEST = "device_text_string_request"
SET_DEVICE_TEXT_STRING = "set_device_text_string"
SET_ALL_LINK_COMMAND_ALIAS = "set_all_link_command_alias"
SET_ALL_LINK = "set_all_link"
ENTER_LINKING_MODE = "enter_linking_mode"
ENTER_UNLINKING_MODE = "enter_unlinking_mode"
GET_INSTEON_ENGINE_VERSION = "get_insteon_engine_version"
PING = "ping"
ID_REQUEST = "id_request"
ID_REQUEST_RESPONSE = "id_request_response"
ON = "on"
ON_FAST = "on_fast"
OFF = "off"
OFF_FAST = "off_fast"
BRIGHTEN_ONE_STEP = "brighten_one_step"
DIM_ONE_STEP = "dim_one_step"
START_MANUAL_CHANGE_DOWN = "start_manual_change_down"
START_MANUAL_CHANGE_UP = "start_manual_change_up"
STOP_MANUAL_CHANGE = "stop_manual_change"
STATUS_REQUEST = "status_request"
GET_OPERATING_FLAGS = "get_operating_flags"
SET_OPERATING_FLAGS = "set_operating_flags"
INSTANT_CHANGE = "instant_change"
MANUALLY_TURNED_OFF = "manually_turned_off"
MANUALLY_TURNED_ON = "manually_turned_on"
REMOTE_SET_BUTTON_TAP1_TAP = "remote_set_button_tap1_tap"
REMOTE_SET_BUTTON_TAP2_TAP = "remote_set_button_tap2_tap"
SET_STATUS = "set_status"
SET_ADDRESS_MSB = "set_address_msb"
POKE_ONE_BYTE = "poke_one_byte"
PEEK_ONE_BYTE = "peek_one_byte"
PEEK_ONE_BYTE_INTERNAL = "peek_one_byte_internal"
POKE_ONE_BYTE_INTERNAL = "poke_one_byte_internal"
ON_AT_RAMP_RATE = "on_at_ramp_rate"
EXTENDED_GET_SET = "extended_get_set"
EXTENDED_GET_RESPONSE = "extended_get_response"
THERMOSTAT_SET_POINT_RESPONSE = "thermostat_set_point_response"
THERMOSTAT_STATUS_RESPONSE = "thermostat_status_response"
EXTENDED_GET_SET_2 = "extended_get_set_2"
OFF_AT_RAMP_RATE = "off_at_ramp_rate"
EXTENDED_READ_WRITE_ALDB = "extended_read_write_aldb"
EXTENDED_TRIGGER_ALL_LINK = "extended_trigger_all_link"
BEEP = "beep"
SPRINKLER_VALVE_ON = "sprinkler_valve_on"
SPRINKLER_VALVE_OFF = "sprinkler_valve_off"
SPRINKLER_PROGRAM_ON = "sprinkler_program_on"
SPRINKLER_PROGRAM_OFF = "sprinkler_program_off"
SPRINKLER_LOAD_INITIALIZATION_VALUES = "sprinkler_load_initialization_values"
SPRINKLER_LOAD_EEPROM_FROM_RAM = "sprinkler_load_eeprom_from_ram"
SPRINKLER_GET_VALVE_STATUS = "sprinkler_get_valve_status"
SPRINKLER_INHIBIT_COMMAND_ACCEPTANCE = "sprinkler_inhibit_command_acceptance"
SPRINKLER_RESUME_COMMAND_ACCEPTANCE = "sprinkler_resume_command_acceptance"
SPRINKLER_SKIP_FORWARD = "sprinkler_skip_forward"
SPRINKLER_SKIP_BACK = "sprinkler_skip_back"
SPRINKLER_ENABLE_PUMP_ON_V8 = "sprinkler_enable_pump_on_v8"
SPRINKLER_DISABLE_PUMP_ON_V8 = "sprinkler_disable_pump_on_v8"
SPRINKLER_BROADCAST_ON = "sprinkler_broadcast_on"
SPRINKLER_BROADCAST_OFF = "sprinkler_broadcast_off"
SPRINKLER_LOAD_RAM_FROM_EEPROM = "sprinkler_load_ram_from_eeprom"
SPRINKLER_SENSOR_ON = "sprinkler_sensor_on"
SPRINKLER_SENSOR_OFF = "sprinkler_sensor_off"
SPRINKLER_DIAGNOSTICS_ON = "sprinkler_diagnostics_on"
SPRINKLER_DIAGNOSTICS_OFF = "sprinkler_diagnostics_off"
SPRINKLER_GET_PROGRAM_REQUEST = "sprinkler_get_program_request"
IO_OUTPUT_ON = "io_output_on"
IO_OUTPUT_OFF = "io_output_off"
IO_ALARM_DATA_REQUEST = "io_alarm_data_request"
IO_WRITE_OUTPUT_PORT = "io_write_output_port"
IO_READ_INPUT_PORT = "io_read_input_port"
IO_GET_SENSOR_VALUE = "io_get_sensor_value"
IO_SET_SENSOR_1_NOMINAL_VALUE = "io_set_sensor_1_nominal_value"
IO_GET_SENSOR_ALARM_DELTA = "io_get_sensor_alarm_delta"
FAN_STATUS_REPORT = "fan_status_report"
IO_WRITE_CONFIGURATION_PORT = "io_write_configuration_port"
IO_READ_CONFIGURATION_PORT = "io_read_configuration_port"
IO_MODULE_LOAD_INITIALIZATION_VALUES = "io_module_load_initialization_values"
IO_MODULE_LOAD_EEPROM_FROM_RAM = "io_module_load_eeprom_from_ram"
IO_MODULE_STATUS_REQUEST = "io_module_status_request"
IO_MODULE_READ_ANALOG_ONCE = "io_module_read_analog_once"
IO_MODULE_READ_ANALOG_ALWAYS = "io_module_read_analog_always"
IO_MODULE_ENABLE_STATUS_CHANGE_MESSAGE = "io_module_enable_status_change_message"
IO_MODULE_DISABLE_STATUS_CHANGE_MESSAGE = "io_module_disable_status_change_message"
IO_MODULE_LOAD_RAM_FROM_EEPROM = "io_module_load_ram_from_eeprom"
IO_MODULE_SENSOR_ON = "io_module_sensor_on"
IO_MODULE_SENSOR_OFF = "io_module_sensor_off"
IO_MODULE_DIAGNOSTICS_ON = "io_module_diagnostics_on"
IO_MODULE_DIAGNOSTICS_OFF = "io_module_diagnostics_off"
POOL_DEVICE_ON = "pool_device_on"
POOL_DEVICE_OFF = "pool_device_off"
POOL_TEMPERATURE_UP = "pool_temperature_up"
POOL_TEMPERATURE_DOWN = "pool_temperature_down"
POOL_LOAD_INITIALIZATION_VALUES = "pool_load_initialization_values"
POOL_LOAD_EEPROM_FROM_RAM = "pool_load_eeprom_from_ram"
POOL_GET_POOL_MODE = "pool_get_pool_mode"
POOL_GET_AMBIENT_TEMPERATURE = "pool_get_ambient_temperature"
POOL_GET_WATER_TEMPERATURE = "pool_get_water_temperature"
POOL_GET_PH = "pool_get_ph"
DOOR_MOVE_RAISE_DOOR = "door_move_raise_door"
DOOR_MOVE_LOWER_DOOR = "door_move_lower_door"
DOOR_MOVE_OPEN_DOOR = "door_move_open_door"
DOOR_MOVE_CLOSE_DOOR = "door_move_close_door"
DOOR_MOVE_STOP_DOOR = "door_move_stop_door"
DOOR_MOVE_SINGLE_DOOR_OPEN = "door_move_single_door_open"
DOOR_MOVE_SINGLE_DOOR_CLOSE = "door_move_single_door_close"
DOOR_STATUS_REPORT_RAISE_DOOR = "door_status_report_raise_door"
DOOR_STATUS_REPORT_LOWER_DOOR = "door_status_report_lower_door"
DOOR_STATUS_REPORT_OPEN_DOOR = "door_status_report_open_door"
DOOR_STATUS_REPORT_CLOSE_DOOR = "door_status_report_close_door"
DOOR_STATUS_REPORT_STOP_DOOR = "door_status_report_stop_door"
DOOR_STATUS_REPORT_SINGLE_DOOR_OPEN = "door_status_report_single_door_open"
DOOR_STATUS_REPORT_SINGLE_DOOR_CLOSE = "door_status_report_single_door_close"
WINDOW_COVERING_OPEN = "window_covering_open"
WINDOW_COVERING_CLOSE = "window_covering_close"
WINDOW_COVERING_STOP = "window_covering_stop"
WINDOW_COVERING_PROGRAM = "window_covering_program"
WINDOW_COVERING_POSITION = "window_covering_position"
THERMOSTAT_TEMPERATURE_UP = "thermostat_temperature_up"
THERMOSTAT_TEMPERATURE_DOWN = "thermostat_temperature_down"
THERMOSTAT_GET_ZONE_INFORMATION = "thermostat_get_zone_information"
THERMOSTAT_CONTROL = "thermostat_control"
THERMOSTAT_SET_COOL_SETPOINT = "thermostat_set_cool_setpoint"
THERMOSTAT_SET_HEAT_SETPOINT = "thermostat_set_heat_setpoint"
THERMOSTAT_EXTENDED_STATUS = "thermostat_extended_status"
THERMOSTAT_TEMPERATURE_STATUS = "thermostat_temperature_status"
THERMOSTAT_HUMIDITY_STATUS = "thermostat_humidity_status"
THERMOSTAT_MODE_STATUS = "thermostat_mode_status"
THERMOSTAT_COOL_SET_POINT_STATUS = "thermostat_cool_set_point_status"
THERMOSTAT_HEAT_SET_POINT_STATUS = "thermostat_heat_set_point_status"
LEAK_DETECTOR_ANNOUNCE = "leak_detector_announce"
ASSIGN_TO_COMPANION_GROUP = "assign_to_companion_group"
SET_SPRINKLER_PROGRAM = "set_sprinkler_program"
SPRINKLER_GET_PROGRAM_RESPONSE = "sprinkler_get_program_response"
IO_SET_SENSOR_NOMINAL_VALUE = "io_set_sensor_nominal_value"
IO_ALARM_DATA_RESPONSE = "io_alarm_data_response"
POOL_SET_DEVICE_TEMPERATURE = "pool_set_device_temperature"
POOL_SET_DEVICE_HYSTERESIS = "pool_set_device_hysteresis"
THERMOSTAT_ZONE_TEMPERATURE_UP = "thermostat_zone_temperature_up"
THERMOSTAT_ZONE_TEMPERATURE_DOWN = "thermostat_zone_temperature_down"
THERMOSTAT_SET_ZONE_HEAT_SETPOINT = "thermostat_set_zone_heat_setpoint"
THERMOSTAT_SET_ZONE_COOL_SETPOINT = "thermostat_set_zone_cool_setpoint"

DEVICE_LINK_CONTROLLER_CREATED = "device_link_controller_created"
DEVICE_LINK_CONTROLLER_REMOVED = "device_link_controller_removed"
DEVICE_LINK_RESPONDER_CREATED = "device_link_responder_created"
DEVICE_LINK_RESPONDER_REMOVED = "device_link_responder_removed"

ALDB_VERSION = "aldb_version"



"""Manages links between devices to identify device state of responders."""
from .. import pub
from ..address import Address
from ..constants import MessageFlagType
from ..topics import DEVICE_LINK_CONTROLLER_CREATED, DEVICE_LINK_RESPONDER_CREATED
from ..utils import subscribe_topic


def _controller_group_topic(responder, group):
    return "{}.{}".format(responder.id, group)


def _topic_to_addr_group(topic):
    elements = topic.name.split(".")
    try:
        address = Address(elements[0])
        group = int(elements[1])
        msg_type = getattr(MessageFlagType, elements[-1].upper())
    except [KeyError, TypeError]:
        return None, None
    return address, group, msg_type


class DeviceLinkManager:
    """Manages links between devices to identify state of responders.

    Listens for broadcast messages from a device and identifies the responders of the device.
    Once the responders are identified, the state of the responders is determined.
    """

    def __init__(self, devices):
        """Init the DeviceLinkManager class."""
        subscribe_topic(self._controller_link_created, DEVICE_LINK_CONTROLLER_CREATED)
        subscribe_topic(self._controller_link_created, DEVICE_LINK_RESPONDER_CREATED)
        self._devices = devices

    def _controller_link_created(self, controller, responder, group):
        controller_group = _controller_group_topic(controller, group)
        subscribe_topic(self._check_responders, controller_group)

    def _check_responders(self, topic=pub.AUTO_TOPIC, **kwargs):
        controller, group, msg_type = _topic_to_addr_group(topic)
        if msg_type != MessageFlagType.ALL_LINK_BROADCAST:
            return
        if group == 0:
            return

        responders = []
        device_c = self._devices[controller]
        for mem_addr in device_c.aldb:
            rec = device_c.aldb[mem_addr]
            if (
                rec.is_in_use
                and rec.is_controller
                and rec.group == group
                and rec.target not in responders
            ):
                responders.append(rec.target)

        for addr in self._devices:
            if (
                addr == self._devices.modem.address
                or addr == device_c.address
                or addr in responders
            ):
                continue
            device_r = self._devices[addr]
            for mem_addr in device_r.aldb:
                rec = device_r.aldb[mem_addr]
                if (
                    rec.is_in_use
                    and rec.is_responder
                    and rec.target == controller
                    and rec.group == group
                ):
                    responders.append(addr)
                    # If the device is a category 1 or 2 device we can pre-load the device state with the
                    # ALDB record data1 field value. We will then check the actual status later.
                    if device_r.cat in [0x01, 0x02] and len(device_r.groups) == 1:
                        device_r.groups[1].value = rec.data1

        for addr in responders:
            device = self._devices[addr]
            if device:
                device.status()



"""Collection of topics mapped to commands (cmd1, cmd2)."""
import logging
from collections import namedtuple
from typing import Iterable

from ..topics import (
    ALL_LINK_CLEANUP_STATUS_REPORT,
    ASSIGN_TO_ALL_LINK_GROUP,
    ASSIGN_TO_COMPANION_GROUP,
    BEEP,
    BRIGHTEN_ONE_STEP,
    DELETE_FROM_ALL_LINK_GROUP,
    DEVICE_TEXT_STRING_REQUEST,
    DIM_ONE_STEP,
    DOOR_MOVE_CLOSE_DOOR,
    DOOR_MOVE_LOWER_DOOR,
    DOOR_MOVE_OPEN_DOOR,
    DOOR_MOVE_RAISE_DOOR,
    DOOR_MOVE_SINGLE_DOOR_CLOSE,
    DOOR_MOVE_SINGLE_DOOR_OPEN,
    DOOR_MOVE_STOP_DOOR,
    DOOR_STATUS_REPORT_CLOSE_DOOR,
    DOOR_STATUS_REPORT_LOWER_DOOR,
    DOOR_STATUS_REPORT_OPEN_DOOR,
    DOOR_STATUS_REPORT_RAISE_DOOR,
    DOOR_STATUS_REPORT_SINGLE_DOOR_CLOSE,
    DOOR_STATUS_REPORT_SINGLE_DOOR_OPEN,
    DOOR_STATUS_REPORT_STOP_DOOR,
    ENTER_LINKING_MODE,
    ENTER_UNLINKING_MODE,
    EXTENDED_GET_RESPONSE,
    EXTENDED_GET_SET,
    EXTENDED_GET_SET_2,
    EXTENDED_READ_WRITE_ALDB,
    EXTENDED_RECEIVED,
    EXTENDED_TRIGGER_ALL_LINK,
    FX_USERNAME,
    GET_INSTEON_ENGINE_VERSION,
    GET_OPERATING_FLAGS,
    ID_REQUEST,
    INSTANT_CHANGE,
    IO_ALARM_DATA_REQUEST,
    IO_ALARM_DATA_RESPONSE,
    IO_GET_SENSOR_ALARM_DELTA,
    IO_GET_SENSOR_VALUE,
    IO_MODULE_DIAGNOSTICS_OFF,
    IO_MODULE_DIAGNOSTICS_ON,
    IO_MODULE_DISABLE_STATUS_CHANGE_MESSAGE,
    IO_MODULE_ENABLE_STATUS_CHANGE_MESSAGE,
    IO_MODULE_LOAD_EEPROM_FROM_RAM,
    IO_MODULE_LOAD_INITIALIZATION_VALUES,
    IO_MODULE_LOAD_RAM_FROM_EEPROM,
    IO_MODULE_READ_ANALOG_ALWAYS,
    IO_MODULE_READ_ANALOG_ONCE,
    IO_MODULE_SENSOR_OFF,
    IO_MODULE_SENSOR_ON,
    IO_MODULE_STATUS_REQUEST,
    IO_OUTPUT_OFF,
    IO_OUTPUT_ON,
    IO_READ_CONFIGURATION_PORT,
    IO_READ_INPUT_PORT,
    IO_SET_SENSOR_1_NOMINAL_VALUE,
    IO_SET_SENSOR_NOMINAL_VALUE,
    IO_WRITE_CONFIGURATION_PORT,
    IO_WRITE_OUTPUT_PORT,
    LEAK_DETECTOR_ANNOUNCE,
    MANUALLY_TURNED_OFF,
    MANUALLY_TURNED_ON,
    OFF,
    OFF_AT_RAMP_RATE,
    OFF_FAST,
    ON,
    ON_AT_RAMP_RATE,
    ON_FAST,
    PEEK_ONE_BYTE,
    PEEK_ONE_BYTE_INTERNAL,
    PING,
    POKE_ONE_BYTE,
    POKE_ONE_BYTE_INTERNAL,
    POOL_DEVICE_OFF,
    POOL_DEVICE_ON,
    POOL_GET_AMBIENT_TEMPERATURE,
    POOL_GET_PH,
    POOL_GET_POOL_MODE,
    POOL_GET_WATER_TEMPERATURE,
    POOL_LOAD_EEPROM_FROM_RAM,
    POOL_LOAD_INITIALIZATION_VALUES,
    POOL_SET_DEVICE_HYSTERESIS,
    POOL_SET_DEVICE_TEMPERATURE,
    POOL_TEMPERATURE_DOWN,
    POOL_TEMPERATURE_UP,
    PRODUCT_DATA_REQUEST,
    REMOTE_SET_BUTTON_TAP1_TAP,
    REMOTE_SET_BUTTON_TAP2_TAP,
    SEND_EXTENDED,
    SEND_STANDARD,
    SET_ADDRESS_MSB,
    SET_ALL_LINK,
    SET_ALL_LINK_COMMAND_ALIAS,
    SET_DEVICE_TEXT_STRING,
    SET_OPERATING_FLAGS,
    SET_SPRINKLER_PROGRAM,
    SET_STATUS,
    SPRINKLER_BROADCAST_OFF,
    SPRINKLER_BROADCAST_ON,
    SPRINKLER_DIAGNOSTICS_OFF,
    SPRINKLER_DIAGNOSTICS_ON,
    SPRINKLER_DISABLE_PUMP_ON_V8,
    SPRINKLER_ENABLE_PUMP_ON_V8,
    SPRINKLER_GET_PROGRAM_REQUEST,
    SPRINKLER_GET_PROGRAM_RESPONSE,
    SPRINKLER_GET_VALVE_STATUS,
    SPRINKLER_INHIBIT_COMMAND_ACCEPTANCE,
    SPRINKLER_LOAD_EEPROM_FROM_RAM,
    SPRINKLER_LOAD_INITIALIZATION_VALUES,
    SPRINKLER_LOAD_RAM_FROM_EEPROM,
    SPRINKLER_PROGRAM_OFF,
    SPRINKLER_PROGRAM_ON,
    SPRINKLER_RESUME_COMMAND_ACCEPTANCE,
    SPRINKLER_SENSOR_OFF,
    SPRINKLER_SENSOR_ON,
    SPRINKLER_SKIP_BACK,
    SPRINKLER_SKIP_FORWARD,
    SPRINKLER_VALVE_OFF,
    SPRINKLER_VALVE_ON,
    STANDARD_RECEIVED,
    START_MANUAL_CHANGE_DOWN,
    START_MANUAL_CHANGE_UP,
    STATUS_REQUEST,
    STOP_MANUAL_CHANGE,
    THERMOSTAT_CONTROL,
    THERMOSTAT_COOL_SET_POINT_STATUS,
    THERMOSTAT_GET_ZONE_INFORMATION,
    THERMOSTAT_HEAT_SET_POINT_STATUS,
    THERMOSTAT_HUMIDITY_STATUS,
    THERMOSTAT_MODE_STATUS,
    THERMOSTAT_SET_COOL_SETPOINT,
    THERMOSTAT_SET_HEAT_SETPOINT,
    THERMOSTAT_SET_POINT_RESPONSE,
    THERMOSTAT_SET_ZONE_COOL_SETPOINT,
    THERMOSTAT_SET_ZONE_HEAT_SETPOINT,
    THERMOSTAT_STATUS_RESPONSE,
    THERMOSTAT_TEMPERATURE_DOWN,
    THERMOSTAT_TEMPERATURE_STATUS,
    THERMOSTAT_TEMPERATURE_UP,
    THERMOSTAT_ZONE_TEMPERATURE_DOWN,
    THERMOSTAT_ZONE_TEMPERATURE_UP,
    WINDOW_COVERING_CLOSE,
    WINDOW_COVERING_OPEN,
    WINDOW_COVERING_POSITION,
    WINDOW_COVERING_PROGRAM,
    WINDOW_COVERING_STOP,
)

Command = namedtuple("Command", "cmd1 cmd2 user_data")
_LOGGER = logging.getLogger(__name__)


def _check_user_data_match(cmd_ud, user_data):
    """Check if the command user_data matches the input user_data."""
    if cmd_ud is None:
        return True
    if not user_data and not cmd_ud:
        return True
    if user_data and not cmd_ud:
        return False
    if cmd_ud and not user_data:
        return False
    for field in cmd_ud:
        if cmd_ud[field] != user_data.get(field):
            return False
    return True


def _check_match(command, cmd1, cmd2, user_data):
    """Check if the current command matches the input values."""
    if command.cmd1 != cmd1:
        return False
    cmd2_match = cmd2 is None or command.cmd2 is None or command.cmd2 == cmd2
    user_data_match = _check_user_data_match(command.user_data, user_data)
    return cmd2_match and user_data_match


class Commands:
    """List of topics and commands."""

    def __init__(self):
        """Init the Commands class."""
        self._topics = {}
        self._commands = {}
        self._use_group = {}

    def add(
        self,
        topic: str,
        cmd1: int,
        cmd2: int,
        user_data: Iterable,
        use_group: bool = False,
    ):
        """Add a command to the list."""
        self._topics[topic] = Command(cmd1, cmd2, user_data)
        self._use_group[topic] = use_group

    def get(self, topic: str):
        """Get the command elements of teh topic."""
        return self._topics.get(topic)

    def get_command(self, topic: str) -> (int, int, bool):
        """Get cmd1 and cmd2 from a topic.

        Returns (cmd1, cmd2, extended)
        """
        return self._topics.get(topic)

    def use_group(self, topic):
        """Return if a topic requires a group number."""
        return self._use_group.get(topic)

    def get_topics(self, cmd1, cmd2, user_data=None, send=False) -> str:
        """Generate a topic from a cmd1, cmd2 and extended flag."""
        found = False
        for topic in self._topics:
            command = self._topics[topic]
            if _check_match(command, cmd1, cmd2, user_data):
                found = True
                yield topic
        if not found:
            if user_data is None:
                yield SEND_STANDARD if send else STANDARD_RECEIVED
            else:
                yield SEND_EXTENDED if send else EXTENDED_RECEIVED


commands = Commands()


commands.add(STANDARD_RECEIVED, -1, None, False)
commands.add(EXTENDED_RECEIVED, -1, None, True)
commands.add(SEND_STANDARD, -2, None, False)
commands.add(SEND_EXTENDED, -2, None, {})
commands.add(ASSIGN_TO_ALL_LINK_GROUP, 0x01, None, False)
commands.add(DELETE_FROM_ALL_LINK_GROUP, 0x02, None, False)
commands.add(PRODUCT_DATA_REQUEST, 0x03, 0x00, None)
commands.add(FX_USERNAME, 0x03, 0x01, False)
commands.add(DEVICE_TEXT_STRING_REQUEST, 0x03, 0x02, False)
commands.add(SET_DEVICE_TEXT_STRING, 0x03, 0x03, {})
commands.add(SET_ALL_LINK_COMMAND_ALIAS, 0x03, 0x04, {})
commands.add(SET_ALL_LINK, 0x03, 0x04, {})
commands.add(ALL_LINK_CLEANUP_STATUS_REPORT, 0x06, None, None, True)
commands.add(ENTER_LINKING_MODE, 0x09, None, None)
commands.add(ENTER_UNLINKING_MODE, 0x0A, None, False)
commands.add(GET_INSTEON_ENGINE_VERSION, 0x0D, None, False)
commands.add(PING, 0x0F, None, False)
commands.add(ID_REQUEST, 0x10, None, False)
commands.add(ON, 0x11, None, None, True)
commands.add(ON_FAST, 0x12, None, None, True)
commands.add(OFF, 0x13, None, None, True)
commands.add(OFF_FAST, 0x14, None, None, True)
commands.add(BRIGHTEN_ONE_STEP, 0x15, None, False, True)
commands.add(DIM_ONE_STEP, 0x16, None, False, True)
commands.add(START_MANUAL_CHANGE_DOWN, 0x17, 0x00, False, True)
commands.add(START_MANUAL_CHANGE_UP, 0x17, 0x01, False, True)
commands.add(STOP_MANUAL_CHANGE, 0x18, None, False, True)
commands.add(STATUS_REQUEST, 0x19, None, None)
commands.add(GET_OPERATING_FLAGS, 0x1F, None, False)
commands.add(SET_OPERATING_FLAGS, 0x20, None, None)
commands.add(INSTANT_CHANGE, 0x21, None, False, True)
commands.add(MANUALLY_TURNED_OFF, 0x22, None, False, True)
commands.add(MANUALLY_TURNED_ON, 0x23, None, False, True)
commands.add(REMOTE_SET_BUTTON_TAP1_TAP, 0x25, 0x01, False)
commands.add(REMOTE_SET_BUTTON_TAP2_TAP, 0x25, 0x02, False)
commands.add(SET_STATUS, 0x27, None, False)
commands.add(SET_ADDRESS_MSB, 0x28, None, False)
commands.add(POKE_ONE_BYTE, 0x29, None, False)
commands.add(PEEK_ONE_BYTE, 0x2B, None, False)
commands.add(PEEK_ONE_BYTE_INTERNAL, 0x2C, None, False)
commands.add(POKE_ONE_BYTE_INTERNAL, 0x2D, None, False)

# cmd2 ne 0x00 => no confict w/ EXTENDED_GET_SET
# Conflict w/ THERMOSTAT_EXTENDED_STATUS but thermostat has no ramp rate so OK
commands.add(ON_AT_RAMP_RATE, 0x2E, None, False, True)
# direct is ed and direct_ack is sd
commands.add(EXTENDED_GET_SET, 0x2E, 0x00, None)
commands.add(EXTENDED_GET_RESPONSE, 0x2E, 0x00, {"d2": 0x01})
commands.add(
    THERMOSTAT_SET_POINT_RESPONSE, 0x2E, 0x00, {"d1": 0x00, "d2": 0x01, "d3": 0x01}
)
commands.add(EXTENDED_GET_SET_2, 0x2E, 0x02, None)
commands.add(THERMOSTAT_STATUS_RESPONSE, 0x2E, 0x02, {"d1": 0x01})

# cmd2 ne 0x00 => no confict w/ read aldb
commands.add(OFF_AT_RAMP_RATE, 0x2F, None, False, True)
# direct is ed and direct_ack is sd
commands.add(EXTENDED_READ_WRITE_ALDB, 0x2F, 0x00, None)
commands.add(EXTENDED_TRIGGER_ALL_LINK, 0x30, None, None)

commands.add(BEEP, 0x30, None, None)
commands.add(SET_SPRINKLER_PROGRAM, 0x40, None, True)
commands.add(SPRINKLER_VALVE_ON, 0x40, None, False)
commands.add(SPRINKLER_GET_PROGRAM_RESPONSE, 0x41, None, True)
commands.add(SPRINKLER_VALVE_OFF, 0x41, None, False)
commands.add(SPRINKLER_PROGRAM_ON, 0x42, None, None)
commands.add(SPRINKLER_PROGRAM_OFF, 0x43, None, None)
commands.add(SPRINKLER_LOAD_INITIALIZATION_VALUES, 0x44, 0x00, False)
commands.add(SPRINKLER_LOAD_EEPROM_FROM_RAM, 0x44, 0x01, False)
commands.add(SPRINKLER_GET_VALVE_STATUS, 0x44, 0x02, False)
commands.add(SPRINKLER_INHIBIT_COMMAND_ACCEPTANCE, 0x44, 0x03, False)
commands.add(SPRINKLER_RESUME_COMMAND_ACCEPTANCE, 0x44, 0x04, False)
commands.add(SPRINKLER_SKIP_FORWARD, 0x44, 0x05, False)
commands.add(SPRINKLER_SKIP_BACK, 0x44, 0x06, False)
commands.add(SPRINKLER_ENABLE_PUMP_ON_V8, 0x44, 0x07, False)
commands.add(SPRINKLER_DISABLE_PUMP_ON_V8, 0x44, 0x08, False)
commands.add(SPRINKLER_BROADCAST_ON, 0x44, 0x09, False)
commands.add(SPRINKLER_BROADCAST_OFF, 0x44, 0x0A, False)
commands.add(SPRINKLER_LOAD_RAM_FROM_EEPROM, 0x44, 0x0B, False)
commands.add(SPRINKLER_SENSOR_ON, 0x44, 0x0C, False)
commands.add(SPRINKLER_SENSOR_OFF, 0x44, 0x0D, False)
commands.add(SPRINKLER_DIAGNOSTICS_ON, 0x44, 0x0E, False)
commands.add(SPRINKLER_DIAGNOSTICS_OFF, 0x44, 0x0F, False)
commands.add(SPRINKLER_GET_PROGRAM_REQUEST, 0x45, None, False)
commands.add(IO_OUTPUT_ON, 0x45, None, False)
commands.add(IO_OUTPUT_OFF, 0x46, None, False)
commands.add(IO_ALARM_DATA_REQUEST, 0x47, 0x00, False)
commands.add(IO_WRITE_OUTPUT_PORT, 0x48, None, False)
commands.add(IO_READ_INPUT_PORT, 0x49, 0x00, False)
commands.add(IO_GET_SENSOR_VALUE, 0x4A, None, False)
commands.add(IO_SET_SENSOR_1_NOMINAL_VALUE, 0x4B, None, False)
commands.add(IO_SET_SENSOR_NOMINAL_VALUE, 0x4B, None, True)
commands.add(IO_GET_SENSOR_ALARM_DELTA, 0x4C, None, False)
commands.add(IO_ALARM_DATA_RESPONSE, 0x4C, 0x00, {})
commands.add(IO_WRITE_CONFIGURATION_PORT, 0x4D, None, False)
commands.add(IO_READ_CONFIGURATION_PORT, 0x4E, 0x00, False)
commands.add(IO_MODULE_LOAD_INITIALIZATION_VALUES, 0x4F, 0x00, False)
commands.add(IO_MODULE_LOAD_EEPROM_FROM_RAM, 0x4F, 0x01, False)
commands.add(IO_MODULE_STATUS_REQUEST, 0x4F, 0x02, False)
commands.add(IO_MODULE_READ_ANALOG_ONCE, 0x4F, 0x03, False)
commands.add(IO_MODULE_READ_ANALOG_ALWAYS, 0x4F, 0x04, False)
commands.add(IO_MODULE_ENABLE_STATUS_CHANGE_MESSAGE, 0x4F, 0x09, False)
commands.add(IO_MODULE_DISABLE_STATUS_CHANGE_MESSAGE, 0x4F, 0x0A, False)
commands.add(IO_MODULE_LOAD_RAM_FROM_EEPROM, 0x4F, 0x0B, False)
commands.add(IO_MODULE_SENSOR_ON, 0x4F, 0x0C, False)
commands.add(IO_MODULE_SENSOR_OFF, 0x4F, 0x0D, False)
commands.add(IO_MODULE_DIAGNOSTICS_ON, 0x4F, 0x0E, False)
commands.add(IO_MODULE_DIAGNOSTICS_OFF, 0x4F, 0x0F, False)
commands.add(POOL_DEVICE_ON, 0x50, None, False)
commands.add(POOL_SET_DEVICE_TEMPERATURE, 0x50, None, {})
commands.add(POOL_DEVICE_OFF, 0x51, None, False)
commands.add(POOL_SET_DEVICE_HYSTERESIS, 0x51, None, {})
commands.add(POOL_TEMPERATURE_UP, 0x52, None, False)
commands.add(POOL_TEMPERATURE_DOWN, 0x53, None, False)
commands.add(POOL_LOAD_INITIALIZATION_VALUES, 0x54, 0x00, False)
commands.add(POOL_LOAD_EEPROM_FROM_RAM, 0x54, 0x01, False)
commands.add(POOL_GET_POOL_MODE, 0x54, 0x02, False)
commands.add(POOL_GET_AMBIENT_TEMPERATURE, 0x54, 0x03, False)
commands.add(POOL_GET_WATER_TEMPERATURE, 0x54, 0x04, False)
commands.add(POOL_GET_PH, 0x54, 0x05, False)
commands.add(DOOR_MOVE_RAISE_DOOR, 0x58, 0x00, False)
commands.add(DOOR_MOVE_LOWER_DOOR, 0x58, 0x01, False)
commands.add(DOOR_MOVE_OPEN_DOOR, 0x58, 0x02, False)
commands.add(DOOR_MOVE_CLOSE_DOOR, 0x58, 0x03, False)
commands.add(DOOR_MOVE_STOP_DOOR, 0x58, 0x04, False)
commands.add(DOOR_MOVE_SINGLE_DOOR_OPEN, 0x58, 0x05, False)
commands.add(DOOR_MOVE_SINGLE_DOOR_CLOSE, 0x58, 0x06, False)
commands.add(DOOR_STATUS_REPORT_RAISE_DOOR, 0x59, 0x00, False)
commands.add(DOOR_STATUS_REPORT_LOWER_DOOR, 0x59, 0x01, False)
commands.add(DOOR_STATUS_REPORT_OPEN_DOOR, 0x59, 0x02, False)
commands.add(DOOR_STATUS_REPORT_CLOSE_DOOR, 0x59, 0x03, False)
commands.add(DOOR_STATUS_REPORT_STOP_DOOR, 0x59, 0x04, False)
commands.add(DOOR_STATUS_REPORT_SINGLE_DOOR_OPEN, 0x59, 0x05, False)
commands.add(DOOR_STATUS_REPORT_SINGLE_DOOR_CLOSE, 0x59, 0x06, False)
commands.add(WINDOW_COVERING_OPEN, 0x60, 0x01, False)
commands.add(WINDOW_COVERING_CLOSE, 0x60, 0x02, False)
commands.add(WINDOW_COVERING_STOP, 0x60, 0x03, False)
commands.add(WINDOW_COVERING_PROGRAM, 0x60, 0x04, False)
commands.add(WINDOW_COVERING_POSITION, 0x61, None, False)
commands.add(THERMOSTAT_TEMPERATURE_UP, 0x68, None, False)
commands.add(THERMOSTAT_ZONE_TEMPERATURE_UP, 0x68, None, {})
commands.add(THERMOSTAT_TEMPERATURE_DOWN, 0x69, None, False)
commands.add(THERMOSTAT_ZONE_TEMPERATURE_DOWN, 0x69, None, {})
commands.add(THERMOSTAT_GET_ZONE_INFORMATION, 0x6A, None, False)
commands.add(THERMOSTAT_CONTROL, 0x6B, None, None, False)
commands.add(THERMOSTAT_SET_COOL_SETPOINT, 0x6C, None, None, False)
commands.add(THERMOSTAT_SET_ZONE_COOL_SETPOINT, 0x6C, None, {})
commands.add(THERMOSTAT_SET_HEAT_SETPOINT, 0x6D, None, None, False)
commands.add(THERMOSTAT_SET_ZONE_HEAT_SETPOINT, 0x6D, None, {})
commands.add(THERMOSTAT_TEMPERATURE_STATUS, 0x6E, None, False)
commands.add(THERMOSTAT_HUMIDITY_STATUS, 0x6F, None, False)
commands.add(THERMOSTAT_MODE_STATUS, 0x70, None, False)
commands.add(THERMOSTAT_COOL_SET_POINT_STATUS, 0x71, None, False)
commands.add(THERMOSTAT_HEAT_SET_POINT_STATUS, 0x72, None, False)
commands.add(LEAK_DETECTOR_ANNOUNCE, 0x70, None, False)
commands.add(ASSIGN_TO_COMPANION_GROUP, 0x81, 0x00, False)



"""Modem command to get the IM configuration."""
import logging

from . import ack_handler
from ..topics import SET_IM_CONFIGURATION
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class SetImConfigurationHandler(OutboundHandlerBase):
    """Handle Set IM Configuration commands."""

    def __init__(self):
        """Init the GetImConfigurationHandler class."""
        super().__init__(topic=SET_IM_CONFIGURATION)

    # pylint: disable=arguments-differ
    def send(
        self,
        disable_auto_linking: bool,
        monitor_mode: bool,
        auto_led: bool,
        deadman: bool,
    ):
        """Send the Set IM Configuration command."""
        super().send(
            disable_auto_linking=disable_auto_linking,
            monitor_mode=monitor_mode,
            auto_led=auto_led,
            deadman=deadman,
        )

    # pylint: disable=arguments-differ
    async def async_send(
        self,
        disable_auto_linking: bool,
        monitor_mode: bool,
        auto_led: bool,
        deadman: bool,
    ):
        """Send the Set IM Configuration command."""
        return await super().async_send(
            disable_auto_linking=disable_auto_linking,
            monitor_mode=monitor_mode,
            auto_led=auto_led,
            deadman=deadman,
        )

    @ack_handler(wait_response=False)
    def handle_ack(self, disable_auto_linking, monitor_mode, auto_led, deadman):
        """Receive the ACK message and return True."""



"""Convert an Insteon Message to a URL for the Hub."""
from ..constants import AllLinkMode
from .messages.outbound import Outbound


def _to_url(host, port, cmd):
    """Convert a host, port and command to a url."""
    return "http://{:s}:{:d}/{:s}".format(host, port, cmd)


def convert_to_url(host, port, msg: Outbound) -> str:
    """Convert a message to a URL."""
    return _to_url(host, port, "3?{:s}=I=3".format(bytes(msg).hex()))


def start_all_link_url(host, port, msg: Outbound):
    """Convert a Start All-Linking message to URL."""
    if msg.mode == AllLinkMode.DELETE:
        return _to_url(host, port, "3?{:s}=I=3".format(bytes(msg).hex()))
    return _to_url(host, port, "0?09{:02d}".format(msg.group))


def cancel_all_linking_url(host, port, msg: Outbound):
    """Convert a Cancel All-Linking message to URL."""
    return _to_url(host, port, "0?0800")



"""Handle All-Link Cleanup Status Reports."""
from ..topics import ALL_LINK_CLEANUP_STATUS_REPORT
from .outbound_base import OutboundHandlerBase
from . import ack_handler, nak_handler


class AllLinkCleanupStatusReport(OutboundHandlerBase):
    """Handle All-Link Cleanup Status Reports."""

    def __init__(self):
        """Init AllLinkCleanupStatusReport class."""
        super().__init__(topic=ALL_LINK_CLEANUP_STATUS_REPORT)

    @ack_handler
    def handle_ack(self):
        """Handle an ACK response."""

    @nak_handler
    def handle_nak(self):
        """Handle a NAK response."""



"""Base class to handle Broadcast messages from devices."""
from datetime import datetime
from ...address import Address
from ...constants import MessageFlagType
from ..inbound_base import InboundHandlerBase


class BroadcastCommandHandlerBase(InboundHandlerBase):
    """Base class to handle inbound Broadcast messages."""

    def __init__(self, topic, address, group=None):
        """Init the BroadcastHandlerBase class."""
        self._address = Address(address)
        self._group = group
        super().__init__(
            topic=topic,
            address=self._address,
            group=self._group,
            message_type=MessageFlagType.ALL_LINK_BROADCAST,
        )
        self._last_command = datetime(1, 1, 1)
        self._last_hops_left = None

    def is_first_message(self, **kwargs):
        """Test if the message is a duplicate."""
        curr_time = datetime.now()
        tdelta = curr_time - self._last_command
        self._last_command = curr_time
        hops_left = kwargs["hops_left"]
        if (
            kwargs["target"].middle != 0
            and hops_left == self._last_hops_left
            and tdelta.seconds < 1
        ):
            return False
        if (
            self._last_hops_left is None
            or (hops_left == self._last_hops_left and tdelta.seconds > 0.7)
            or hops_left > self._last_hops_left
            or tdelta.seconds >= 2
        ):
            self._last_hops_left = hops_left
            return True
        return False



"""Receive an X10 message."""

from .. import inbound_handler
from ... import pub
from ...constants import X10Commands, X10CommandType
from ...topics import X10_RECEIVED
from ...utils import byte_to_command, byte_to_unitcode, parse_x10
from ...x10_address import create
from ..inbound_base import InboundHandlerBase


class X10Received(InboundHandlerBase):
    """Receive an X10 message."""

    def __init__(self):
        """Init the X10Received class."""
        super().__init__(topic=X10_RECEIVED)
        self._last_housecode = None
        self._last_unitcode = None

    @inbound_handler
    def handle_x10_received(self, raw_x10, x10_flag):
        """Manage X10 inbound messages."""
        housecode, uc_or_cmd = parse_x10(raw_x10)
        if x10_flag == X10CommandType.COMMAND:
            self._notify_subscribers(housecode, uc_or_cmd)
        else:
            self._last_housecode = housecode
            self._last_unitcode = byte_to_unitcode(uc_or_cmd)

    def _notify_subscribers(self, housecode, uc_or_cmd):
        """Notify subscribes that a command was received."""
        cmd = byte_to_command(uc_or_cmd)
        topic = ""
        if cmd in [
            X10Commands.ALL_LIGHTS_OFF,
            X10Commands.ALL_LIGHTS_ON,
            X10Commands.ALL_UNITS_OFF,
        ]:
            topic = "x10{}.{}".format(housecode.lower(), str(cmd))
        if self._last_housecode == housecode and self._last_unitcode is not None:
            address = create(housecode, self._last_unitcode)
            topic = "{}.{}".format(address.id, str(cmd))
        if topic:
            pub.sendMessage(topic)
            self._last_housecode = None
            self._last_unitcode = None



"""Container for a default link record."""
from collections import namedtuple

DefaultLink = namedtuple(
    "DefaultLink",
    "is_controller group dev_data1 dev_data2 dev_data3 "
    "modem_data1 modem_data2 modem_data3",
)



"""Manage outbound ON command to a device."""
from .. import status_handler, status_ack_handler
from ... import pub
from ...topics import STATUS_REQUEST
from .direct_command import DirectCommandHandlerBase


class StatusRequestCommand(DirectCommandHandlerBase):
    """Manage an outbound Status command to a device.

    TODO Confirm that the status command is best with a single status handler
    rather than a handler per status command (ie. one for state 1 and one for
    state 2)
    """

    _status_type = None
    _status_active = False

    def __init__(self, address, status_type: int = 0):
        """Init the OnLevelCommand class."""
        super().__init__(topic=STATUS_REQUEST, address=address)
        self._status_type = status_type
        if status_type:
            self._subscriber_topic = f"{self._subscriber_topic}_{status_type}"

    @property
    def status_type(self):
        """Return the type of status message."""
        return self._status_type

    @property
    def status_active(self):
        """Return if the status command is active."""
        return self._status_active

    @status_active.setter
    def status_active(self, value: bool):
        """Set if the status command is active."""
        self._status_active = bool(value)

    # pylint: disable=arguments-differ, useless-super-delegation
    async def async_send(self):
        """Send the ON command async."""
        return await super().async_send(status_type=self._status_type)

    @status_ack_handler()
    def handle_ack(self, cmd1, cmd2, user_data):
        """Handle the message ACK."""

    @status_handler
    def handle_direct_ack(self, topic=pub.AUTO_TOPIC, **kwargs):
        """Handle the ON response direct ACK."""
        cmd1 = kwargs.get("cmd1")
        cmd2 = kwargs.get("cmd2")
        if cmd2 is not None:
            self._call_subscribers(db_version=cmd1, status=cmd2)



"""Thermostat temperature up command."""
from .. import direct_ack_handler
from ...topics import THERMOSTAT_CONTROL
from ...constants import ThermostatMode
from .direct_command import DirectCommandHandlerBase


class ThermostatModeCommand(DirectCommandHandlerBase):
    """Manage an outbound THERMOSTAT_TEMPERATURE_DOWN command to a device."""

    def __init__(self, address):
        """Init the TemperatureUpCommand class."""
        super().__init__(topic=THERMOSTAT_CONTROL, address=address)

    # pylint: disable=arguments-differ
    def send(self, mode):
        """Send the OFF command."""
        super().send(mode=mode)

    # pylint: disable=arguments-differ
    async def async_send(self, mode):
        """Send the OFF command async."""
        if mode == ThermostatMode.HEAT:
            send_mode = 0x04
        elif mode == ThermostatMode.COOL:
            send_mode = 0x05
        elif mode == ThermostatMode.AUTO:
            send_mode = 0x06
        elif mode == ThermostatMode.FAN_ALWAYS_ON:
            send_mode = 0x07
        elif mode == ThermostatMode.FAN_AUTO:
            send_mode = 0x08
        elif mode == ThermostatMode.OFF:
            send_mode = 0x09
        return await super().async_send(mode=send_mode)

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the OFF response direct ACK."""
        if cmd2 == 0x04:
            mode = ThermostatMode.HEAT
        elif cmd2 == 0x05:
            mode = ThermostatMode.COOL
        elif cmd2 == 0x06:
            mode = ThermostatMode.AUTO
        elif cmd2 == 0x07:
            mode = ThermostatMode.FAN_ALWAYS_ON
        elif cmd2 == 0x08:
            mode = ThermostatMode.FAN_AUTO
        elif cmd2 == 0x09:
            mode = ThermostatMode.OFF

        self._call_subscribers(mode=mode)



"""Advanced ALDB tools."""
import json
from binascii import unhexlify
from os import path

import aiofiles

from .. import devices
from ..constants import ALDBStatus, LinkStatus, ResponseStatus
from ..managers.link_manager import find_broken_links
from ..managers.saved_devices_manager import aldb_rec_to_dict, dict_to_aldb_record
from .tools_base import ToolsBase
from ..aldb.aldb_record import ALDBRecord


def _convert_val(val):
    if '"' not in val and "'" not in val:
        try:
            return int(val)
        except ValueError:
            try:
                return float(val)
            except ValueError:
                val = val.strip('"').strip("'")
    else:
        val = val.strip('"').strip("'")
    if val in ["y", "c"]:
        return True
    if val == ["n", "r"]:
        return False
    return val


def _parse_record(*args):
    try:
        address = args[0]
        if not devices[address]:
            raise IndexError
    except IndexError:
        return None, None, {}

    try:
        rec_id = args[1]
        mem_addr = int.from_bytes(unhexlify(rec_id), byteorder="big")
    except IndexError:
        return address, None, {}
    except ValueError:
        return address, -1, {}

    kwargs = {}
    for arg in args[2:]:
        kwarg, val = arg.split("=")
        val = _convert_val(val)
        kwargs[kwarg] = val

    return address, mem_addr, kwargs


class AdvancedTools(ToolsBase):
    """Advanced ALDB tools."""

    async def do_print_aldb(self, *args, **kwargs):
        """Print the records in an All-Link Database.

        Usage:
            print_aldb <ADDRESS>|all
        """
        await self._print_aldb(*args)

    async def do_add_link(self, *args, **kwargs):
        """Add a link to a device All-Link Database.

        For modems use the add_im_link command.

        Usage:
            add_device_link address group target c|r [data1 data2 data3]

        address: Address of the device to add the link [0 - 255]
        group: Group number of the link [0 - 255]
        target: The device target the link refers to
        c|r: Controller or responder link type (c=Controller, r=Responder)
        """

        args = args[0].split()
        address = None
        group = None
        target = None
        controller = None
        data1 = None
        data2 = None
        data3 = None
        try:
            address = args[0]
            group = int(args[1])
            target = args[2]
            if args[3].lower() in ["r", "c"]:
                controller = args[3].lower() == "c"
            else:
                controller = None
            data1 = int(args[4])
            data2 = int(args[5])
            data3 = int(args[6])
        except (IndexError, ValueError):
            pass

        address = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )
        if not address:
            return
        device = devices[address]
        if device.aldb.status != ALDBStatus.LOADED:
            self._log_stdout(
                f"The All-Link Database for device {device.address} must be loaded first."
            )
            return

        if not group:
            group = await self._get_int("Group number", values=range(0, 256))

        target = await self._get_addresses(
            address=target,
            allow_cancel=True,
            allow_all=False,
            prompt="Enter link target address",
        )
        if not target:
            return
        if controller is None:
            controller = bool(
                await self._get_char(
                    "Controller or responder (c=Controller, r=Responder)",
                    values=["c", "r"],
                )
                == "c"
            )
        if data1 is None:
            data1 = await self._get_int("Data 1", default=0, values=range(0, 256))
        if data2 is None:
            data2 = await self._get_int("Data 2", default=0, values=range(0, 256))
        if data3 is None:
            data3 = await self._get_int("Data 3", default=0, values=range(0, 256))

        c_r = "c" if controller else "r"
        self._log_command(
            f"add_device_link {address} {group} {target} {c_r} {data1} {data2} {data3}"
        )

        device.aldb.add(
            group=group,
            target=target,
            controller=controller,
            data1=data1,
            data2=data2,
            data3=data3,
        )
        result = await device.aldb.async_write()
        if device.is_battery:
            self._log_command(
                f"Device {device.address} is battery operated. The record will be written when the device wakes up."
            )
        elif result != ResponseStatus.SUCCESS:
            self._log_stdout(
                f"An issue occured writing to the database of device {device.address}"
            )
        else:
            self._log_stdout(
                f"The record was successfully writen to device {device.address}"
            )

    async def do_remove_link(self, *args, **kwargs):
        """Remove a link from the All-Link Database.

        Usage:
            remove_link address mem_addr

            address: Address of the device
            mem_addr: Memory address of the link to remove (i.e. 0f7f)
        """
        args = args[0].split()
        address = None
        mem_addr = None
        try:
            address = args[0]
            mem_addr_str = args[1]
            mem_addr = int.from_bytes(unhexlify(mem_addr_str), byteorder="big")
        except (IndexError, ValueError):
            pass

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return
        device = devices[addresses[0]]
        if device.aldb.status != ALDBStatus.LOADED:
            self._log_stdout(
                f"The All-Link Database for device {device.address} must be loaded first."
            )
            return

        if mem_addr is None:
            mem_addr_str = await self._get_int(
                "Memory address of the record to remove (Be very careful!)"
            )
            if mem_addr == "":
                return
            mem_addr = int.from_bytes(unhexlify(mem_addr_str), byteorder="big")

        if device.aldb[mem_addr] is None:
            self._log_stdout(
                f"Record {mem_addr:04x} was not found in the ALDB of device {device.address}."
            )
            return

        self._log_command(f"remove_link {address} {mem_addr:04x}")
        device.aldb.remove(mem_addr)
        result = await device.aldb.async_write()
        if device.is_battery:
            self._log_command(
                f"Device {device.address} is battery operated. The record will be removed when the device wakes up."
            )
        elif result != ResponseStatus.SUCCESS:
            self._log_stdout(
                f"An issue occured writing to the database of device {device.address}"
            )
        else:
            self._log_stdout(
                f"The record was successfully removed from the device {device.address}"
            )

    async def do_add_im_link(self, *args, **kwargs):
        """Add a link to the Insteon Modem (IM)."""
        args = args[0].split()
        group = None
        target = None
        controller = None
        data1 = None
        data2 = None
        data3 = None
        try:
            group = int(args[0])
            target = args[1]
            if args[3].lower() in ["r", "c"]:
                controller = args[2].lower() == "c"
            else:
                controller = None
            data1 = int(args[3])
            data2 = int(args[4])
            data3 = int(args[5])
        except (IndexError, ValueError):
            pass

        if devices.modem.aldb.status != ALDBStatus.LOADED:
            self._log_stdout(
                f"The All-Link Database for the Insteon Modem (IM) must be loaded first."
            )
            return

        if not group:
            group = await self._get_int("Group number", values=range(0, 256))

        target = await self._get_addresses(
            address=target,
            allow_cancel=True,
            allow_all=False,
            prompt="Enter link target address",
        )
        if not target:
            return
        if controller is None:
            controller = bool(
                await self._get_char(
                    "Controller or responder (c=Controller, r=Responder)",
                    values=["c", "r"],
                )
                == "c"
            )
        if data1 is None:
            data1 = await self._get_int("Data 1", default=0, values=range(0, 256))
        if data2 is None:
            data2 = await self._get_int("Data 2", default=0, values=range(0, 256))
        if data3 is None:
            data3 = await self._get_int("Data 3", default=0, values=range(0, 256))

        c_r = "c" if controller else "r"
        self._log_command(f"add_im_link {group} {target} {c_r} {data1} {data2} {data3}")

        devices.modem.aldb.add(
            group=group,
            target=target,
            controller=controller,
            data1=data1,
            data2=data2,
            data3=data3,
        )
        result = await devices.modem.aldb.async_write()
        if result != ResponseStatus.SUCCESS:
            self._log_stdout(
                f"An issue occured writing to the database of the Insteon Modem (IM)."
            )
        else:
            self._log_stdout(
                f"The record was successfully writen to the Insteon Modem (IM)."
            )

    async def do_remove_im_link(self, *args, **kwargs):
        """Remove a link from the Insteon Modem (IM)."""
        args = args[0].split()
        group = None
        target = None
        controller = None
        try:
            group = int(args[0])
            target = args[1]
            if args[3].lower() in ["r", "c"]:
                controller = args[2].lower() == "c"
            else:
                controller = None
        except (IndexError, ValueError):
            pass

        if devices.modem.aldb.status != ALDBStatus.LOADED:
            self._log_stdout(
                f"The All-Link Database for the Insteon Modem (IM) must be loaded first."
            )
            return

        if not group:
            group = await self._get_int("Group number", values=range(0, 256))

        target = await self._get_addresses(
            address=target,
            allow_cancel=True,
            allow_all=False,
            prompt="Enter link target address",
        )
        if not target:
            return
        if controller is None:
            controller = bool(
                await self._get_char(
                    "Controller or responder (c=Controller, r=Responder)",
                    values=["c", "r"],
                )
                == "c"
            )

        c_r = "c" if controller else "r"
        self._log_command(f"remove_im_link {group} {target} {c_r}")

        devices.modem.aldb.remove(group=group, target=target, controller=controller)
        result = await devices.modem.aldb.async_write()
        if result != ResponseStatus.SUCCESS:
            self._log_stdout(
                f"An issue occured writing to the database of the Insteon Modem (IM)."
            )
        else:
            self._log_stdout(
                f"The record was successfully writen to the Insteon Modem (IM)."
            )

    def do_find_broken_links(self, *args, **kwargs):
        """Find broken links between devices."""
        broken_links = find_broken_links(devices)
        self._log_stdout("Device   Mem Addr Target    Group Mode Status")
        self._log_stdout(
            "-------- -------- --------- ----- ---- ----------------------------------------"
        )
        for address in broken_links:
            for mem_addr in broken_links[address]:
                rec, status = broken_links[address][mem_addr]
                if status == LinkStatus.MISSING_CONTROLLER:
                    status_txt = "Missing controller"
                elif status == LinkStatus.MISSING_RESPONDER:
                    status_txt = "Missing responder"
                elif status == LinkStatus.MISSING_TARGET:
                    status_txt = "Target device not found"
                elif status == LinkStatus.TARGET_DB_NOT_LOADED:
                    status_txt = "Cannot verify - Target ALDB not loaded"
                if rec.is_controller:
                    mode = "C"
                else:
                    mode = "R"
                self._log_stdout(
                    f"{address:s}     {mem_addr:04x} {rec.target:s} {rec.group:5d}   {mode:s} {status_txt:.40s}"
                )

    async def do_change_link(self, *args, **kwargs):
        """Add a link to a device All-Link Database.

        For modems use the add_im_link command.

        Usage:
            add_device_link address rec_id field=value [field=value field=value]


        address: Address of the device to add the link [0 - 255]
        rec_id: ALDB record id (i.e. 1fff. Can be found with the print_aldb command)

        Field can be any of the following:
            in_use: Should the record be in use
                y: record is active
                n: record is inactive (same as deleting the record)
            mode: Controller or responder link type
                c: Controller
                r: Responder)
            group: Group number of the link [0 - 255]
            target: The address of the device target the link refers to
            data1: Data field 1 [0 - 255]
            data2: Data field 2 [0 - 255]
            data3: Data feild 3 [0 - 255]
            force: Force the write to happen (overrides ALDB load status)
        """

        args = args[0].split()
        address, mem_addr, kwargs = _parse_record(*args)
        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        if mem_addr is None:
            rec_id = await self._get_char("Record ID (eg. 1fff)")
            try:
                mem_addr = int.from_bytes(unhexlify(rec_id), byteorder="big")
            except ValueError:
                self._log_stdout("Invalid record id")

        device = devices[address]
        rec = device.aldb[mem_addr]
        if not rec:
            self._log_stdout(f"All-Link record not found")
            return

        if not kwargs:
            while True:
                arg = await self._get_char(
                    "Enter argument (eg. mode=c). Press enter for last argument",
                    default="",
                )
                if arg == "":
                    break
                kwarg, val = arg.split("=")
                val = _convert_val(val)
                kwargs[kwarg] = val

        for kwarg in kwargs:
            if kwarg not in [
                "in_use",
                "mode",
                "target",
                "group",
                "data1",
                "data2",
                "data3",
            ]:
                self._log_stdout(f"Invalid argument: {kwarg}")
                return

        in_use = rec.is_in_use if not kwargs.get("in_use") else kwargs.get("in_use")
        controller = rec.is_controller if not kwargs.get("mode") else kwargs.get("mode")
        target = rec.target if not kwargs.get("target") else kwargs.get("target")
        group = rec.group if not kwargs.get("group") else kwargs.get("group")
        data1 = rec.data1 if not kwargs.get("data1") else kwargs.get("data1")
        data2 = rec.data2 if not kwargs.get("data2") else kwargs.get("data2")
        data3 = rec.data3 if not kwargs.get("data3") else kwargs.get("data3")
        force = kwargs.get("force", False)

        device.aldb.modify(
            mem_addr=mem_addr,
            group=group,
            target=target,
            in_use=in_use,
            controller=controller,
            data1=data1,
            data2=data2,
            data3=data3,
        )
        result = await device.aldb.async_write(force=force)
        if device.is_battery:
            self._log_command(
                f"Device {device.address} is battery operated. The record will be written when the device wakes up."
            )
        elif result != ResponseStatus.SUCCESS:
            self._log_stdout(
                f"An issue occured writing to the database of device {device.address}"
            )
        else:
            self._log_stdout(
                f"The record was successfully writen to device {device.address}"
            )

    async def do_export_aldb(self, *args, **kwargs):
        """Export the All-Link Database for a device to a file.

        Usage:
            export_aldb address location [filename]

        address: Address of the device to export
        location: Directory location of where to place the file (use `.` for the current directory).
        filename: Filename of the device. The default file name is address_aldb.json
        """
        args = args[0].split()
        location = None
        filename = None
        try:
            address = args[0]
            try:
                location = args[1]
                filename = args[2]
            except IndexError:
                pass
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        address = addresses[0]
        device = devices[address]

        if location is None:
            location = self._get_workdir()

        if filename is None:
            filename = f"{device.address.id}_aldb.json"
            filename = await self._get_char("Filename", default=filename)

        aldb_dict = {}
        for mem in device.aldb:
            rec = device.aldb[mem]
            rec_dict = aldb_rec_to_dict(rec)
            aldb_dict[mem] = rec_dict
        await self._write_aldb_file(aldb_dict, location, filename)

    async def do_replace_aldb(self, *args, **kwargs):
        """Read and replace the device ALDB with the contents of a file.

        Reads a JSON formated file with All-Link Database records and replaces
        the current records in the device. To create the JSON file use the
        `export_aldb` command.

        WARNING: This method is very dangerous and can make your device non-responsive.

        Usage:
            replace_aldb address location [filename]

        address: Address of the device
        location: Directory location of the file (use `.` for the current directory).
        filename: Name of the file to use (default is address_aldb.json)
        """
        args = args[0].split()
        location = None
        filename = None
        try:
            address = args[0]
            try:
                location = args[1]
                filename = args[2]
            except IndexError:
                pass
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )

        if not addresses:
            return

        address = addresses[0]
        device = devices[address]

        if location is None:
            location = self._get_workdir()

        if filename is None:
            filename = f"{device.address.id}_aldb.json"
            filename = await self._get_char("Filename", default=filename)

        aldb_dict = await self._read_device_aldb(location, filename)
        aldb_recs = dict_to_aldb_record(aldb_dict)
        device.aldb.load_saved_records(ALDBStatus.LOADED, aldb_recs)
        last_rec = None
        for mem in device.aldb:
            rec = device.aldb[mem]
            device.aldb[mem] = rec
            last_rec = rec
        if last_rec and not last_rec.is_high_water_mark:
            mem = last_rec.mem_addr - 8
            new_rec = ALDBRecord(mem, False, 0, "00.00.00", 0, 0, 0, False, True, 0, 0)
            device.aldb[mem] = new_rec
        self._log_stdout("Preview of changes:")
        self._print_aldb_out([device.address])
        confirm = await self._get_char("Continue", default="n", values=["y", "n"])
        if confirm == "y":
            done, not_done = await device.aldb.async_write(force=True)

        if not done:
            self._log_stdout("No records were written")
        elif not_done:
            self._log_stdout("Some records not written")
        else:
            self._log_stdout("All records written succesfully")

    async def _write_aldb_file(self, aldb_dict, location, filename):
        device_file = path.join(location, filename)
        try:
            async with aiofiles.open(device_file, "w") as afp:
                out_json = json.dumps(aldb_dict, indent=2)
                await afp.write(out_json)
                await afp.flush()
        except FileNotFoundError as ex:
            self._log_stdout(f"Cannot write to file {device_file}")
            self._log_stdout(f"Exception: {str(ex)}")

    async def _read_device_aldb(self, location, filename):
        """Load device information from the device info file."""
        aldb_dict = []
        try:
            device_file = path.join(location, filename)
            async with aiofiles.open(device_file, "r") as afp:
                json_file = ""
                json_file = await afp.read()
            try:
                aldb_dict = json.loads(json_file)
            except json.decoder.JSONDecodeError:
                self._log_stdout("Loading ALDB file failed")
        except FileNotFoundError:
            self._log_stdout("ALDB file not found")
        return aldb_dict



"""Handle sending a read request for ALDB records."""
import logging

from ...address import Address
from ...topics import SET_OPERATING_FLAGS
from .direct_command import DirectCommandHandlerBase
from .. import direct_nak_handler
from ...constants import ResponseStatus

_LOGGER = logging.getLogger(__name__)


class SetOperatingFlagsCommand(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=SET_OPERATING_FLAGS, address=address)

    # pylint: disable=arguments-differ
    def send(self, cmd: int, extended=False):
        """Send Get Operating Flags message."""
        super().send(cmd=cmd, extended=extended)

    # pylint: disable=arguments-differ
    async def async_send(self, cmd: int, extended=False):
        """Send Get Operating Flags message asyncronously."""
        cmd_response = await super().async_send(cmd=cmd, extended=extended)
        if cmd_response == ResponseStatus.UNCLEAR and not extended:
            _LOGGER.debug("Attempting resend with extended message")
            cmd_response = await super().async_send(cmd=cmd, extended=True)
        return cmd_response

    @direct_nak_handler
    def handle_direct_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the direct ACK."""
        self._call_subscribers(response=cmd2)



"""Manage inbound OFF All-Link Cleanup command to a device."""

from .. import inbound_handler
from ...address import Address
from ...topics import ON
from .all_link_cleanup_command import AllLinkCleanupCommandHandlerBase


class OnAllLinkCleanupInbound(AllLinkCleanupCommandHandlerBase):
    """Off All-Link Cleanup command inbound."""

    def __init__(self, address, group):
        """Init the OffAllLinkCleanupAckInbound class."""
        self._address = Address(address)
        self._group = group
        super().__init__(topic=ON, address=self._address, group=self._group)

    @inbound_handler
    def handle_command(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the ON All-Link Cleanup from a device."""
        self._call_subscribers()



"""Operating flag or Extended Property for all device types."""
from .subscriber_base import SubscriberBase


class DeviceFlagBase(SubscriberBase):
    """Operating flag or Extended Property."""

    def __init__(self, topic, name, flag_type: type, is_reversed=False):
        """Init the DeviceFlag class."""
        super().__init__(topic)
        self._name = name
        self._value = None
        self._type = flag_type
        self._new_value = None
        self._is_dirty = False
        self._is_loaded = False
        self._reversed = is_reversed

    @property
    def name(self):
        """Return the flag name."""
        return self._name

    @property
    def value(self):
        """Return the value of the flag."""
        return self._value

    @property
    def new_value(self):
        """Return the new value of the flag."""
        return self._new_value

    @new_value.setter
    def new_value(self, value):
        """Set the new value of the flag.

        This is the primary method to set the value of the flag.
        It sets the `new_value` property and the `is_dirty` property.
        """
        if value != self._value:
            self._new_value = self._type(value)
            self._is_dirty = True

    @property
    def is_dirty(self):
        """Return if the Operating flag has been changed."""
        return self._is_dirty

    @property
    def is_loaded(self):
        """Return if the Operating flag has been loaded."""
        return self._is_loaded

    @property
    def is_reversed(self):
        """Return if the device flag is reverse of this value."""
        return self._reversed

    def load(self, value):
        """Load the flag from the device value.

        Only use this method to update the value of the flag from the value
        of the device.

        This method updates the `is_loaded` property and clears the `new value` and
        `is_dirty` properties.
        """
        if self._type is bool and self._reversed:
            value = not value
        update_value = self._type(value) if value is not None else self._type(0)
        self._is_dirty = False
        self._new_value = None
        self._is_loaded = True
        is_changed = update_value != self._value
        self._value = update_value
        if is_changed:
            self._call_subscribers(name=self._name, value=self._value)



"""Insteon inbound message data structure defintion."""
import logging
from binascii import hexlify

from . import MessageBase
from ...constants import MessageId
from .message_definition import MessageDefinition
from .message_definitions import FLD_EXT_SEND_ACK, FLD_STD_SEND_ACK, INBOUND_MSG_DEF
from .message_flags import MessageFlags

_LOGGER = logging.getLogger(__name__)


def trim_data(raw_data: bytearray):
    """Trim bad data from the front of a message byte stream."""
    data_bytes = bytes(raw_data)
    while data_bytes and data_bytes[0] != 0x02:
        data_bytes = data_bytes[1:]
    return data_bytes


class Inbound(MessageBase):
    """Insteon inbound message data definition."""

    def __init__(self, msg_def: MessageDefinition, raw_data: bytearray):
        """Init the Inbound message class."""
        self._fields = msg_def.fields
        self._len = len(msg_def)

        slices = self._create_slices()
        field_vals = self._slice_data(slices, raw_data)
        super().__init__(msg_def, **field_vals)

    def __len__(self):
        """Emit the length of the message."""
        return self._len

    def _create_slices(self):
        slices = []
        slices.append(slice(0, 1))
        slices.append(slice(1, 2))
        start = 2
        for field in self._fields:
            stop = start + field.length
            if field.length > 1:
                slices.append(slice(start, stop))
            else:
                slices.append(start)
            start = stop
        return slices

    def _slice_data(self, slices, raw_data):
        curr_slice = 2
        field_vals = {}
        for field in self._fields:
            val = field.type(raw_data[slices[curr_slice]])
            field_vals[field.name] = val
            curr_slice += 1
        return field_vals

    def __str__(self):
        """Emit the message in hex."""
        return hexlify(bytes(self)).decode()


def create(raw_data: bytearray) -> (Inbound, bytearray):
    """Create a message from a raw byte array."""

    def _remaining_data(msg, raw_data):
        if msg is not None:
            return raw_data[len(msg) :]
        return raw_data

    def _create_message(msg_def, raw_data):
        msg_len = 2
        for field in msg_def.fields:
            msg_len += field.length
        if len(raw_data) >= msg_len:
            msg = Inbound(msg_def, raw_data)
            return msg, _remaining_data(msg, raw_data)
        _LOGGER.debug("Message too short")
        return None, raw_data

    def _standard_message(raw_data):
        flag_byte = 5
        flags = MessageFlags(raw_data[flag_byte])
        if flags.is_extended:
            msg_def = MessageDefinition(MessageId.SEND_EXTENDED, FLD_EXT_SEND_ACK)
            msg, remaining_data = _create_message(msg_def, raw_data)
        else:
            msg_def = MessageDefinition(MessageId.SEND_STANDARD, FLD_STD_SEND_ACK)
            msg, remaining_data = _create_message(msg_def, raw_data)
        return msg, remaining_data

    _LOGGER.debug("IN CREATE: %s", raw_data.hex())
    data_bytes = trim_data(raw_data)
    if len(data_bytes) < 2:
        _LOGGER.debug("Message less than 2 bytes")
        return None, raw_data
    try:
        msg_id = MessageId(data_bytes[1])
        msg_def = INBOUND_MSG_DEF.get(msg_id)
        if msg_def is not None:
            if len(data_bytes) < len(msg_def):
                _LOGGER.debug("Full message not received")
                return None, raw_data
            if msg_def.message_id == MessageId.SEND_STANDARD:
                msg, remaining_data = _standard_message(data_bytes)
            else:
                msg, remaining_data = _create_message(msg_def, data_bytes)
            return msg, bytearray(remaining_data)
    except ValueError:
        _LOGGER.debug("Message ID not found: 0x%02x", data_bytes[1])
        _LOGGER.debug("Bad Data: %s", raw_data.hex())
        truncate = 1 if data_bytes[1] == 0x02 else 2
        return None, bytearray(data_bytes[truncate:])

    return None, raw_data



"""Direct commands from the Insteom Modem to the device.

These messages represent commands sent to the device where the Insteon
Modem is the controller and the device is the responder in the
All-Link database.

Direct commands are sent to the device to trigger an All-Link group and/or
a state change on the device.
"""
